import unittest
from typing import TYPE_CHECKING

from pydistsim import NetworkGenerator, Simulation
from pydistsim.demo_algorithms.santoro2007.traversal import DFT, DFStar
from pydistsim.demo_algorithms.santoro2007.yoyo import YoYo
from pydistsim.network.rangenetwork import BidirectionalRangeNetwork
from pydistsim.utils.testing import PyDistSimTestCase

if TYPE_CHECKING:
    from pydistsim.network.node import Node


class TestYoYo(PyDistSimTestCase):

    def test_santoro2007(self):
        node_range = 100
        nets = [
            [(100, 100)],
            [(100, 100), (175, 250), (250, 175), (100, 250), (175, 175), (100, 175)],
            [
                (100, 100),
                (150, 200),
                (175, 175),
                (175, 100),
                (250, 175),
                (250, 250),
                (325, 250),
                (325, 325),
                (325, 400),
            ],
        ]

        for i, node_positions in enumerate(nets, start=1):
            with self.subTest(i=i):
                net = BidirectionalRangeNetwork()
                for node_pos in node_positions:
                    net.add_node(pos=node_pos, commRange=node_range)

                name = "Special %d" % i

                sim = Simulation(net)
                sim.algorithms = (YoYo,)

                sim.run(1)

                sim.algorithms[0].check_algorithm_initialization()

                sim.run(100_000)

                sim.algorithms[0].check_algorithm_termination()

                min_id = min(sim.network.nodes(), key=lambda node: node._internal_id)._internal_id
                for node in sim.network.nodes():
                    if node._internal_id != min_id:
                        # Check if every other node is PRUNED
                        assert node.status == YoYo.Status.PRUNED, "%s: Node %d has status %s, not PRUNED" % (
                            name,
                            node._internal_id,
                            node.status,
                        )
                    else:
                        # Check if the node with the smallest ID is the LEADER
                        assert node.status == YoYo.Status.LEADER, "%s: Node %d has status %s, not LEADER" % (
                            name,
                            node._internal_id,
                            node.status,
                        )

    def test_santoro2007_random(self):
        N_ITERS = 1
        N_NETWORKS = 15
        N_NODES_STEP = 5

        for i in range(N_ITERS):
            for n_nodes in range(N_NODES_STEP, N_NETWORKS * N_NODES_STEP + N_NODES_STEP, N_NODES_STEP):
                with self.subTest(i=i, n_nodes=n_nodes):
                    net_gen = NetworkGenerator(n_nodes, directed=False)
                    net = net_gen.generate_random_network()

                    name = " %d, %d nodes" % (i, n_nodes)

                    sim = Simulation(net)
                    sim.algorithms = (YoYo,)

                    sim.run(1)

                    sim.algorithms[0].check_restrictions()
                    sim.algorithms[0].check_algorithm_initialization()

                    sim.run(100_000)

                    sim.algorithms[0].check_algorithm_termination()

                    min_id = min(sim.network.nodes(), key=lambda node: node._internal_id)._internal_id
                    for node in sim.network.nodes():
                        if node._internal_id == min_id:
                            # Check if the node with the smallest ID is the LEADER
                            assert node.status == YoYo.Status.LEADER, "%s: Node %d has status %s, not LEADER" % (
                                name,
                                node._internal_id,
                                node.status,
                            )
                        else:
                            # Check if every other node is PRUNED
                            assert node.status == YoYo.Status.PRUNED, "%s: Node %d has status %s, not PRUNED" % (
                                name,
                                node._internal_id,
                                node.status,
                            )


class TestTraversalDFT(unittest.TestCase):

    def setUp(self):
        net_gen = NetworkGenerator(100, directed=False)
        self.net = net_gen.generate_random_network()

        self.visited = []

        # Asigna el algoritmo

        def visitedAction(node: "Node"):
            self.visited.append(node.unbox()._internal_id)

        self.algorithms = ((DFT, {"visitedAction": visitedAction}),)

    def test_traversal(self):
        sim = Simulation(self.net)
        sim.algorithms = self.algorithms

        sim.run(100_000)

        for node in self.net.nodes():
            assert node.status == DFT.Status.DONE, "Node %d is not DONE" % node._internal_id

        assert sorted(node._internal_id for node in self.net.nodes()) == sorted(self.visited)


class TestTraversalDFStar(PyDistSimTestCase):

    def setUp(self):
        net_gen = NetworkGenerator(100, directed=False)
        self.net = net_gen.generate_random_network()

        self.visited = []

        # Asigna el algoritmo

        def visitedAction(node: "Node"):
            self.visited.append(node.unbox()._internal_id)

        self.algorithms = ((DFStar, {"visitedAction": visitedAction}),)

    def test_traversal(self):
        sim = Simulation(self.net)
        sim.algorithms = self.algorithms

        sim.run(100_000)

        for node in self.net.nodes():
            assert node.status == DFStar.Status.DONE, "Node %d is not DONE" % node._internal_id

        assert sorted(node._internal_id for node in self.net.nodes()) == sorted(self.visited)



from numpy import array, average, diag, dot, linalg, ones, sqrt

from pydistsim.demo_algorithms.niculescu2003.floodingupdate import FloodingUpdate


class Trilaterate(FloodingUpdate):

    required_params = (
        # key in memory for true position data (only landmarks)
        "truePositionKey",
        # key in memory for storing estimated position
        "positionKey",
        # key in memory for storing hopsize data
        "hopsizeKey",
    )

    def initiator_condition(self, node):
        return node.memory[self.truePositionKey] is not None  # if landmark

    def initiator_data(self, node):
        return node.memory[self.hopsizeKey]

    def handle_flood_message(self, node, message):
        if self.hopsizeKey in node.memory:
            return None
        node.memory[self.hopsizeKey] = message.data
        self.estimate_position(node)
        return node.memory[self.hopsizeKey]

    def estimate_position(self, node):
        TRESHOLD = 0.1
        MAX_ITER = 10

        # get landmarks with hopsize data
        landmarks = list(node.memory[self.dataKey].keys())
        # calculate estimated distances
        if len(landmarks) >= 3:
            landmark_distances = [node.memory[self.dataKey][lm][2] * node.memory[self.hopsizeKey] for lm in landmarks]
            landmark_positions = [array(node.memory[self.dataKey][lm][:2]) for lm in landmarks]
            # take centroid as initial estimation
            pos = average(landmark_positions, axis=0)
            W = diag(ones(len(landmarks)))
            counter = 0

            def dist(x, y):
                return sqrt(dot(x - y, x - y))

            while True:
                J = array([(lp - pos) / dist(lp, pos) for lp in landmark_positions])
                range_correction = array(
                    [dist(landmark_positions[li], pos) - landmark_distances[li] for li, lm in enumerate(landmarks)]
                )
                pos_correction = dot(linalg.inv(dot(dot(J.T, W), J)), dot(dot(J.T, W), range_correction))
                pos = pos + pos_correction
                counter += 1
                if sqrt(sum(pos_correction**2)) < TRESHOLD or counter >= MAX_ITER:
                    break
            if counter < MAX_ITER:
                node.memory[self.positionKey] = pos



# flake8: noqa: E731

import unittest
from typing import TYPE_CHECKING

from pydistsim.message import Message
from pydistsim.network import NetworkGenerator
from pydistsim.network.behavior import NetworkBehaviorModel

if TYPE_CHECKING:
    from pydistsim.network.network import NetworkType
    from pydistsim.network.node import Node

CANT_MESSAGES = 100


# all messages are sent
def all_sent(source: "Node", dest: "Node", network: "NetworkType"):
    return len(source.outbox) == 0


# some but not all messages are received
def only_some_received(source: "Node", dest: "Node", network: "NetworkType"):
    return len(dest.inbox) > 0 and len(dest.inbox) < CANT_MESSAGES


# all messages are received
def all_received(source: "Node", dest: "Node", network: "NetworkType"):
    return len(dest.inbox) == CANT_MESSAGES


def none_received(source: "Node", dest: "Node", network: "NetworkType"):
    return len(dest.inbox) == 0


# inbox is ordered by data in message
def assert_order(source: "Node", dest: "Node", network: "NetworkType"):
    return dest.inbox == sorted(dest.inbox, key=(lambda x: x.data), reverse=True)


# inbox is unordered
def assert_unordered(source: "Node", dest: "Node", network: "NetworkType"):
    return dest.inbox != sorted(dest.inbox, key=(lambda x: x.data), reverse=True)


# none in transit
def none_in_transit(source: "Node", dest: "Node", network: "NetworkType"):
    return len(network.get_transit_messages(source, dest)) + len(network.get_transit_messages(dest, source)) == 0


def all_in_transit(source: "Node", dest: "Node", network: "NetworkType"):
    return len(network.get_transit_messages(source, dest)) == CANT_MESSAGES


# some in transit
def some_in_transit(source: "Node", dest: "Node", network: "NetworkType"):
    return len(network.get_transit_messages(source, dest)) > 0


def delay_only_first_message(network: "NetworkType", message: "Message"):
    if hasattr(network, "first_message_sent") and network.first_message_sent:
        return 0
    network.first_message_sent = True
    return 100


DelayOnlyFirstMessage = NetworkBehaviorModel(
    message_ordering=True,
    message_delay_indicator=delay_only_first_message,
    bounded_communication_delays=True,
    message_loss_indicator=None,
    clock_increment=None,
)


class TestBehaviorModel(unittest.TestCase):
    TESTS = {
        NetworkBehaviorModel.LikelyRandomLossCommunication: (
            "LikelyRandomLossCommunication",
            [all_sent, only_some_received, assert_order, none_in_transit],
        ),
        NetworkBehaviorModel.IdealCommunication: (
            "IdealCommunication",
            [all_sent, all_received, assert_order, none_in_transit],
        ),
        NetworkBehaviorModel.UnorderedCommunication: (
            "UnorderedCommunication",
            [all_sent, all_received, assert_unordered, none_in_transit],
        ),
        NetworkBehaviorModel.RandomDelayCommunication: (
            "RandomDelayCommunication",
            [all_sent, assert_order, some_in_transit],
        ),
        NetworkBehaviorModel.UnorderedRandomDelayCommunication: (
            "UnorderedRandomDelayCommunication",
            [all_sent, only_some_received, assert_unordered, some_in_transit],
        ),
        DelayOnlyFirstMessage: (
            "DelayOnlyFirstMessage",
            [all_sent, none_received, assert_order, all_in_transit],
        ),
    }

    def test(self):
        for comm_props, (name, tests) in self.TESTS.items():
            with self.subTest(comm_props=comm_props):
                net = NetworkGenerator(2, enforce_connected=True).generate_random_network()

                node_source = net.nodes_sorted()[0]
                node_dest = net.nodes_sorted()[1]
                for i in range(1, CANT_MESSAGES + 1):
                    message = Message(
                        source=node_source,
                        destination=node_dest,
                        data=i,
                    )
                    node_source.push_to_outbox(message, destination=node_dest)

                print(f"Running tests for {name}")
                net.behavioral_properties = comm_props
                net.communicate()

                def data_messages(messages):
                    return tuple(map(lambda x: x.data, messages))

                # Prints for debugging

                print(f"node_dest.outbox={data_messages(node_dest.outbox)}")
                print(f"node_dest.inbox={data_messages(node_dest.inbox)}")
                print(f"transit_messages={data_messages(net.get_transit_messages(node_source, node_dest))}")
                print(f"lost_messages={data_messages(net.get_lost_messages(node_source, node_dest))}")

                print(f"{len(node_dest.outbox)=}")
                print(f"{len(node_dest.inbox)=}")
                print(f"{len(net.get_transit_messages(node_source, node_dest))=}")
                print(f"{len(net.get_lost_messages(node_source, node_dest))=}")

                for test in tests:
                    assert test(node_source, node_dest, net)



from copy import copy, deepcopy
from inspect import getmembers
from typing import TYPE_CHECKING

from pydistsim.logging import logger
from pydistsim.observers import ObserverManagerMixin
from pydistsim.restrictions import CheckableRestriction
from pydistsim.restrictions.base_restriction import ApplicableRestriction

if TYPE_CHECKING:
    from pydistsim.network import NetworkType
    from pydistsim.simulation import Simulation


class AlgorithmMeta(type):
    """Metaclass for required and default params extension and update."""

    def __new__(cls, clsname, bases, dct):
        """
        Collect required and default params from class bases and extend and
        update those in dct that have been sent for new class.

        """
        rps = []
        dps = {}
        for base in bases[::-1]:
            base_rps = dict(getmembers(base)).get("required_params", [])
            rps.extend(base_rps)
            base_dps = dict(getmembers(base)).get("default_params", {})
            dps.update(base_dps)
        rps.extend(dct.get("required_params", []))
        dps.update(dct.get("default_params", {}))
        all_params = rps + list(dps.keys())

        assert len(rps) == len(set(rps)), "Some required params %s defined in multiple classes." % str(rps)
        assert len(all_params) == len(
            set(all_params)
        ), "Required params {} and default params {} should be unique.".format(
            str(rps),
            str(list(dps.keys())),
        )

        dct["required_params"] = tuple(rps)
        dct["default_params"] = dps

        logger.trace(f"Checking for __configure_class__ in dct of {clsname}")
        __configure_class__ = None
        if "__configure_class__" in dct:
            logger.trace("Found __configure_class__ in dct")
            __configure_class__ = dct["__configure_class__"]
        else:
            logger.trace("Did not find __configure_class__ in dct")
            for base in bases:
                if hasattr(base, "__configure_class__"):
                    logger.trace(f"Found __configure_class__ in base {base.__name__}")
                    __configure_class__ = base.__configure_class__
                    break
        if __configure_class__:
            logger.trace(f"Calling __configure_class__ for {clsname}")
            __configure_class__(clsname, bases, dct)

        return super().__new__(cls, clsname, bases, dct)


class BaseAlgorithm(ObserverManagerMixin, metaclass=AlgorithmMeta):
    """
    Abstract base class for all algorithms.

    Currently there are two main subclasses:
    * NodeAlgorithm used for distributed algorithms
    * NetworkAlgorithm used for centralized algorithms

    When writing new algorithms make them subclass either of NodeAlgorithm or
    NetworkAlgorithm.

    Every algorithm instance has a set of required and default params:
    * Required params must be given to algorithm initializer as a keyword
        arguments.
    * Default params can be given to algorithm initializer as a keyword
        arguments, if not their class defines default value.

    Note: On algorithm initialization all params are converted to instance
    attributes.

    For example:

    class SomeAlgorithm(NodeAlgorithm):
        required_params = ('rp1',)
        default_params = {'dp1': 'dv1',}

    >>> net = Network()
    >>> alg = SomeAlgorithm(net, rp1='rv1')
    >>> alg.rp1
    'rv1'
    >>> alg.dp1
    'dv1'

    Params in algorithm subclasses are inherited from its base classes, that
    is, required params are extended and default are updated:
    * required_params are union of all required params of their ancestor
      classes.
    * default_params are updated so default values are overridden in
      subclasses
    """

    required_params = ()
    default_params = {}

    algorithm_restrictions = ()
    "Tuple of restrictions that must be satisfied for the algorithm to run."

    def __init__(self, simulation: "Simulation", **kwargs):
        super().__init__()
        self.simulation: "Simulation" = simulation
        self.name = self.__class__.__name__
        logger.trace("Instance of {} class has been initialized.", self.name)

        for required_param in self.required_params:
            if required_param not in list(kwargs.keys()):
                raise AlgorithmException(f"Missing required param {required_param} for algorithm {self.name}.")

        # set default params
        for dp, val in list(self.default_params.items()):
            self.__setattr__(dp, val)

        # override default params
        for kw, arg in list(kwargs.items()):
            self.__setattr__(kw, arg)

    def __eq__(self, value: object) -> bool:
        return self.__dict__ == value.__dict__ and isinstance(value, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"

    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]

        # Shallow copy of the object
        copy_a = copy(self)
        memo[id(self)] = copy_a

        # Deep copy of the mutable attributes
        copy_a.simulation = deepcopy(self.simulation, memo)

        return copy_a

    @property
    def network(self) -> "NetworkType":
        return self.simulation.network

    def step(self, check_restrictions: bool, step: int):
        raise NotImplementedError

    def is_initialized(self):
        return self.simulation.algorithmState["step"] != 1 and self.simulation.get_current_algorithm() == self

    def is_halted(self):
        """
        Check if the distributed algorithm has come to an end or deadlock.
        """
        raise NotImplementedError

    def check_restrictions(self):
        """
        Check if the restrictions are satisfied. Does not apply ApplicableRestrictions.
        """
        logger.debug("Checking restrictions for algorithm {}.", self.name)
        for restriction in self.algorithm_restrictions:
            if issubclass(restriction, CheckableRestriction):
                logger.debug("Checking restriction {}.", restriction.__name__)
                if not restriction.check(self.network):
                    raise AlgorithmException(
                        f"Restriction {restriction.__name__} not satisfied for this network.\n"
                        + restriction.get_explanation(self.network)
                    )

    def apply_restrictions(self):
        """
        Apply all applicable restrictions.
        """
        for restriction in self.algorithm_restrictions:
            if issubclass(restriction, ApplicableRestriction):
                restriction.apply(self.network)

    def reset(self):
        """
        Reset the algorithm to its initial state.
        """
        ...


class AlgorithmException(Exception):
    pass



"""
This module contains classes for representing networks and generating networks.
"""

# flake8: noqa: F401

from pydistsim.network.generator import NetworkGenerator, NetworkGeneratorException
from pydistsim.network.network import (
    BidirectionalNetwork,
    Network,
    NetworkException,
    NetworkType,
)
from pydistsim.network.node import Node
from pydistsim.network.rangenetwork import (
    BidirectionalRangeNetwork,
    CompleteRangeType,
    RangeNetwork,
    RangeNetworkType,
    RangeType,
    SquareDiscRangeType,
    UdgRangeType,
)
from pydistsim.simulation import AlgorithmsParam



"""
This module contains the definition of the axioms of the simulation environment.

These axioms are the basis of the definition of the distributed computing environment. They are modeled in the
implementation itself so there is no need to implement the check method.
"""

from typing import TYPE_CHECKING

from pydistsim.restrictions.base_restriction import CheckableRestriction

if TYPE_CHECKING:
    from pydistsim.network.network import NetworkType

from abc import ABC


class SimulationAxiom(CheckableRestriction, ABC):
    "Base class for simulation axioms."

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        """
        Check if the axiom is satisfied by the network.
        As the axioms are modeled in the implementation itself,
        it is not necessary to implement this method.

        :param network: The network object representing the simulation network.
        :type network: NetworkType
        :return: True if the axiom is satisfied, False otherwise.
        :rtype: bool
        """
        return True


class FiniteCommunicationDelays(SimulationAxiom):
    "In the absence of failures, communication delays are finite."


class LocalOrientation(SimulationAxiom):
    "An entity can distinguish among its in and out neighbors."



"""
Set of utils that operate on network nodes when they have defined tree
in their memory under tree_key key.
tree_key -- key in node memory (dictionary) where parent and
           children data is stored in format:
           {'parent': parent_node,
            'children': [child_node1, child_node2 ...]}
"""

from typing import TYPE_CHECKING, Union

from pydistsim.algorithm.node_wrapper import NeighborLabel, NodeAccess

if TYPE_CHECKING:
    from pydistsim.network.network import NetworkType
    from pydistsim.network.node import Node

NodeType = Union["Node", "NodeAccess", "NeighborLabel"]


def get_root_node(net: "NetworkType", tree_key: str = "mst") -> NodeType:
    """
    Return root node in network tree.
    """
    check_tree_key(net, tree_key)

    node = net.nodes_sorted()[0]
    while node.memory[tree_key]["parent"] and node.memory[tree_key]["parent"].unbox() in net.nodes():
        node = node.memory[tree_key]["parent"].unbox()
    return node


def check_tree_key(net: "NetworkType", tree_key: str):
    if len(net.nodes()) == 0:
        raise TreeNetworkException("Network has no nodes.")

    for node in net.nodes():
        if tree_key not in node.memory:
            raise Missingtree_key(tree_key)


class TreeNetworkException(Exception):
    pass


class Missingtree_key(TreeNetworkException):
    def __init__(self, tree_key):
        self.tree_key = tree_key

    def __str__(self):
        return "At least one node is missing '%s' key in memory." % self.tree_key



"""
This module contains the classes for observers and the mixin class for objects that can have observers.
"""

from typing import TYPE_CHECKING

from pydistsim.logging import logger

if TYPE_CHECKING:
    from pydistsim.algorithm import BaseAlgorithm
    from pydistsim.message import Message
    from pydistsim.network import Node
    from pydistsim.simulation import Simulation


class ObservableEvents(str):
    """
    Enum-like class for the events that can be observed.
    :class:`StrEnum` from the :mod:`enum` module is not used because it does not support inheritance.
    """

    added = "added"
    step_done = "step_done"
    message_sent = "message_sent"
    sim_state_changed = "sim_state_changed"
    algorithm_started = "algorithm_started"
    algorithm_finished = "algorithm_finished"
    network_changed = "network_changed"
    message_delivered = "message_delivered"
    node_status_changed = "node_status_changed"

    _all = (
        added,
        step_done,
        message_sent,
        sim_state_changed,
        algorithm_started,
        algorithm_finished,
        network_changed,
        message_delivered,
        node_status_changed,
    )


class Observer:
    """
    Base class for observers.

    Subclasses must implement the `on_{event_name}` methods for the events they want to listen to.

    If an observer is notified of an event it does not listen to, the notification is discarded
    and a debug message is logged.

    To define the events an observer listens to, set the `events` attribute to a list of event names.
    The definitive list of allowed events is the union of the `events` attribute of every class in
    the inheritance chain.
    For more information, refer to the :meth:`__get_allowed_events__` method.

    Example of concrete implementations are:
    - :class:`MetricCollector` class in :mod:`pydistsim.metrics`
    - :class:`QThreadObserver` class in :mod:`pydistsim.gui.simulationgui`
    """

    events = [ObservableEvents.added]

    @classmethod
    def __get_allowed_events__(cls) -> set[ObservableEvents]:
        def get_bases(cls):
            for base_cls in cls.__bases__:
                yield from get_bases(base_cls)
                yield base_cls

        allowed_events = set(cls.events)
        for base_cls in get_bases(cls):
            if issubclass(base_cls, Observer) or base_cls == Observer:
                allowed_events.update(base_cls.events)
        return allowed_events

    def notify(self, event: ObservableEvents, *args, **kwargs):
        allowed_events = self.__get_allowed_events__()
        if event not in allowed_events:
            logger.trace(
                "Invalid event name '{}' for observer type {}. Valid events are: {}.",
                event,
                self.__class__.__name__,
                ", ".join(allowed_events),
            )
            return
        getattr(self, f"on_{event.lower()}")(*args, **kwargs)

    def on_added(self, observable: "ObserverManagerMixin") -> None:
        """
        Called when the observer is added to an observable object.

        :param observable: The observable object to which the observer is added.
        :type observable: ObserverManagerMixin

        :return: None
        """
        ...


class AlgorithmObserver(Observer):
    """
    Observer for algorithm events.
    """

    events = [
        ObservableEvents.step_done,
        ObservableEvents.algorithm_started,
    ]

    def on_step_done(self, algorithm: "BaseAlgorithm") -> None: ...

    def on_algorithm_started(self, algorithm: "BaseAlgorithm") -> None: ...


class SimulationObserver(Observer):
    """
    Observer for simulation events.
    """

    events = [
        ObservableEvents.sim_state_changed,
        ObservableEvents.algorithm_finished,
        ObservableEvents.network_changed,
    ]

    def on_sim_state_changed(self, simulation: "Simulation") -> None: ...

    def on_algorithm_finished(self, algorithm: "BaseAlgorithm") -> None: ...

    def on_network_changed(self, simulation: "Simulation") -> None: ...


class NetworkObserver(Observer):
    """
    Observer for network events.
    No such events are defined yet.
    """

    events = []


class NodeObserver(Observer):
    """
    Observer for node events.
    No such events are defined yet.
    """

    events = [
        ObservableEvents.node_status_changed,
        ObservableEvents.message_delivered,
        ObservableEvents.message_sent,
    ]

    def on_node_status_changed(self, node: "Node", previous_status: str, new_status: str) -> None: ...

    def on_message_delivered(self, message: "Message") -> None: ...

    def on_message_sent(self, message: "Message") -> None: ...


class NodeNetworkObserver(NodeObserver, NetworkObserver):
    """
    Observer for node AND network events.
    """


class ObserverManagerMixin:
    """
    Mixin class for objects that can have observers.

    Subclasses must call the `super().__init__()` method in their :meth:`__init__` method or
    add the :attr:`observers` attribute.
    """

    def __init__(self, *args, **kwargs):
        self.observers: set[Observer] = set()
        super().__init__(*args, **kwargs)

    def add_observers(self, *observers: "Observer"):
        for observer in observers:
            self.observers.add(observer)
            observer.notify(ObservableEvents.added, self)

    def clear_observers(self):
        self.observers = set()

    def notify_observers(self, event: ObservableEvents, *args, **kwargs):
        for observer in self.observers:
            observer.notify(event, *args, **kwargs)



"""
This module contains the :class:`AlgorithmBenchmark` class, which is used to benchmark the performance of a given
algorithm.
"""

from collections.abc import Callable, Iterable
from random import randint, random
from time import time
from typing import Any, Literal

from pandas import DataFrame

from pydistsim.algorithm.node_algorithm import AlgorithmException, NodeAlgorithm
from pydistsim.logging import logger
from pydistsim.metrics import MetricCollector
from pydistsim.network.behavior import NetworkBehaviorModel
from pydistsim.network.generator import NetworkGenerator
from pydistsim.network.network import BidirectionalNetwork, Network, NetworkType
from pydistsim.simulation import AlgorithmsParam, Simulation


def decide_directed_class(directed: bool):
    if directed:
        return Network
    else:
        return BidirectionalNetwork


def decide_rect_shape(n: int):
    "Auxiliary function to choose shape so that `a * b ≈ n` and neither a or b are equal to 1 (most of the time)."

    b = max(randint(int(n / 6), int(n / 3)), 1)
    a = int(n / b)
    return {"a": a, "b": b}


DETERMINISTIC_GENERATORS = {
    "complete": lambda n, directed: NetworkGenerator.generate_complete_network(n, decide_directed_class(directed)),
    "hypercube": lambda n, directed: NetworkGenerator.generate_hypercube_network(
        n, network_type=decide_directed_class(directed)
    ),
    "ring": lambda n, directed: NetworkGenerator.generate_ring_network(n, decide_directed_class(directed)),
    "star": lambda n, directed: NetworkGenerator.generate_star_network(n, decide_directed_class(directed)),
    "square mesh": lambda n, directed: NetworkGenerator.generate_mesh_network(
        n, torus=False, network_type=decide_directed_class(directed)
    ),
    "square torus": lambda n, directed: NetworkGenerator.generate_mesh_network(
        n, torus=True, network_type=decide_directed_class(directed)
    ),
}


RANDOM_GENERATORS = {
    "homogeneous": lambda n, directed: NetworkGenerator(n, directed=directed).generate_homogeneous_network(random()),
    "rectangular mesh": lambda n, directed: NetworkGenerator.generate_mesh_network(
        torus=False, **decide_rect_shape(n), network_type=decide_directed_class(directed)
    ),
    "rectangular torus": lambda n, directed: NetworkGenerator.generate_mesh_network(
        torus=True, **decide_rect_shape(n), network_type=decide_directed_class(directed)
    ),
}

GENERATORS = {"DETERMINISTIC": DETERMINISTIC_GENERATORS, "RANDOM": RANDOM_GENERATORS}


class AlgorithmBenchmark:
    """
    This class is used to benchmark the performance of a given algorithm.


    ### Usage

    .. code-block:: python

        from pydistsim.benchmark import AlgorithmBenchmark
        from pydistsim.network.behavior import NetworkBehaviorModel

        benchmark = AlgorithmBenchmark(
            ((Flood, {"initial_information": "Hello Wold!"}), ),
            network_sizes=range(1, 40),
            network_behavior=NetworkBehaviorModel.UnorderedRandomDelayCommunication,
        )

        benchmark.run()

        plot = benchmark.plot_analysis() # In IPython, this will create a plot with the results

        df = benchmark.get_results_dataframe() # Get the results as a pandas DataFrame

    ### Params

    :param algorithm: The algorithm to benchmark. It can be a single algorithm or a tuple of algorithms.
    :type algorithm: AlgorithmsParam =  tuple[type[BaseAlgorithm] | tuple[type[BaseAlgorithm], dict]]
    :param max_time: The maximum time in seconds to run the benchmark.
    :type max_time: float, optional
    :param network_sizes: The sizes of the networks to generate and run the algorithm on.
    :type network_sizes: Iterable[int], optional
    :param directed_network: If the generated networks must be directed.
    :type directed_network: bool, False by default
    :param check_algorithm_termination: Whether to check if the algorithm terminated correctly. Only for NodeAlgorithms.
    :type check_algorithm_termination: bool, optional
    :param network_behavior: The network behavior model to use.
    :type network_behavior: NetworkBehaviorModel, optional
    :param metric_collector_factory: A factory function to create a MetricCollector instance.
    :type metric_collector_factory: Callable[[], MetricCollector], optional
    :param network_generators: A dict of dicts containing network generators.
    :type network_generators: dict[Literal["DETERMINISTIC", "RANDOM"], dict[str, Callable[[int, bool], "NetworkType"]]]
    :param network_repeat_count: A dict mapping generation type with the amount of repeats.
    :type network_repeat_count: dict[Literal["DETERMINISTIC", "RANDOM"], int]
    """

    def __init__(
        self,
        algorithm: AlgorithmsParam,
        max_time: float = float("inf"),
        network_sizes: Iterable[int] = range(1, 20),
        directed_network: bool = False,
        check_algorithm_termination: bool = True,
        network_behavior: "NetworkBehaviorModel" = NetworkBehaviorModel.IdealCommunication,
        metric_collector_factory: Callable[[], MetricCollector] = MetricCollector,
        network_generators: dict[
            Literal["DETERMINISTIC", "RANDOM"], dict[str, Callable[[int], "NetworkType"]]
        ] = GENERATORS,
        network_repeat_count: dict[Literal["DETERMINISTIC", "RANDOM"], int] = {"DETERMINISTIC": 1, "RANDOM": 5},
    ):
        self.results = []
        self.errors = []

        self.algorithm = algorithm
        self.max_time = max_time
        self.network_sizes = list(network_sizes)
        self.directed_network = directed_network
        self.check_algorithm_termination = check_algorithm_termination
        self.network_behavior = network_behavior
        self.metric_collector_factory = metric_collector_factory
        self.network_generators = network_generators
        self.network_repeat_count = network_repeat_count

    def run(self):
        """
        Run the benchmark
        """
        start_time = time()

        tests = self.get_test_params()

        for network_size, gen_name, gen in tests:
            if time() - start_time > self.max_time:
                logger.warning(
                    f"Time limit reached, stopping benchmark before on {gen_name} network of size {network_size}"
                )
                return

            try:
                logger.trace(f"Generating network '{gen_name}' of size {network_size}")
                network: "NetworkType" = gen(network_size, self.directed_network)
            except ValueError as e:
                logger.debug(f"Failed to generate '{gen_name}' network of size {network_size}. Reason: {e}")
                continue

            logger.trace(f"Running simulation on '{gen_name}' network of size {len(network)}")
            network.behavioral_properties = self.network_behavior

            metrics = self.metric_collector_factory()
            sim = Simulation(network, self.algorithm)
            algorithms_instances: Iterable[NodeAlgorithm] = sim._algorithms
            sim.add_observers(metrics)

            try:
                sim.run()
            except Exception as e:
                self._save_error(gen_name, network, e, "Algorithm failed to run.")

            if self.check_algorithm_termination:
                try:
                    for algorithm_instance in algorithms_instances:
                        algorithm_instance.check_algorithm_termination()
                except AlgorithmException as e:
                    self._save_error(gen_name, network, e, "S_term check failed.")

            self._save_result(gen_name, network, **metrics.create_report())

    def get_test_params(self) -> Iterable[tuple[int, str, Callable[[int], "NetworkType"]]]:
        tests = (
            (network_size, generation_name, generator)
            for network_size in self.network_sizes
            for gen_type, gen_dict in self.network_generators.items()
            for i in range(self.network_repeat_count[gen_type])
            for generation_name, generator in gen_dict.items()
        )

        return tests

    def _save_result(self, gen_name: str, network: "NetworkType", **metrics):
        logger.debug(f"Algorithm terminated correctly on '{gen_name}' network of size {len(network)}")
        self.results.append(
            {
                "Net. node count": len(network),
                "Net. edge count": len(network.edges()),
                "Network type": gen_name,
                **metrics,
            }
        )

    def _save_error(self, gen_name: str, network: "NetworkType", e: Exception, error_type: str):
        logger.error(f"{error_type} On '{gen_name}' network of size {len(network)}")
        self.errors.append(
            {
                "Net. node count": len(network),
                "Net. edge count": len(network.edges()),
                "Network type": gen_name,
                "Error type": error_type,
                "exception_args": e.args,
            }
        )

    def get_results_dataframe(self, grouped: bool = False):
        """
        Get the results of the benchmark as a pandas DataFrame.

        :param grouped: Group results by size and generation type.
        :param grouped: bool, optional
        :return: Results of the benchmark as a table.
        :rtype: DataFrame
        """

        # Group by "Net. node count", "Network type", "Net. edge count" and calculate the mean of the other columns
        # and return without the grouped rows
        if not grouped:
            return DataFrame(self.results)

        return (
            DataFrame(self.results).groupby(["Net. node count", "Network type", "Net. edge count"]).mean().reset_index()
        )

    def plot_analysis(
        self,
        x_vars: Iterable[str] = None,
        y_vars: Iterable[str] = None,
        result_filter: Callable[[dict[str, Any]], bool] = None,
        grouped=True,
        pairplot_kwargs: dict[str, Any] = {},
    ):
        """
        Plot the results of the benchmark using seaborn pairplot.

        ### Example call:

        .. code-block:: python

            benchmark.plot_analysis(
                x_vars=["Net. node count"],
                y_vars=["Qty. of messages sent", "Qty. of steps"],
                result_filter=lambda df: df["Network type"] in ('complete', 'ring'),
            )

        ### Params

        :param x_vars: The variables to plot on the x-axis. By default, it is node count and edge count.
        :type x_vars: Iterable[str], optional
        :param y_vars: The variables on the y-axis. By default, it is all variables except node count and edge count.
        :type y_vars: Iterable[str], optional
        :param result_filter: Filter the results row by row. Optional.
        :type result_filter: Callable[[dict[str, Any]], bool]
        :param grouped: Whether to group the results by network size and generation type.
        :type grouped: bool, optional
        :return: The seaborn pairplot object.
        """
        import seaborn as sns

        df = self.get_results_dataframe(grouped)

        if y_vars is None:
            y_vars = set(df.columns)
            y_vars.remove("Network type")
        else:
            y_vars = set(y_vars)

        x_vars = x_vars or {"Net. node count", "Net. edge count"}
        y_vars.remove("Net. node count") if "Net. node count" in y_vars else None
        y_vars.remove("Net. edge count") if "Net. edge count" in y_vars else None

        if result_filter:
            df = df[df.apply(result_filter, axis=1)]

        hue = None
        if df["Network type"].nunique() > 1:
            hue = "Network type"

        return sns.pairplot(df, x_vars=x_vars, y_vars=y_vars, hue=hue, diag_kind="hist", **pairplot_kwargs)



"""
Restrictions relating to `a priori` knowledge of the network.
"""

from abc import ABC
from typing import TYPE_CHECKING

from pydistsim.restrictions.base_restriction import (
    ApplicableRestriction,
    CheckableRestriction,
)

if TYPE_CHECKING:
    from pydistsim.network.network import NetworkType


class KnowledgeRestriction(CheckableRestriction, ABC):
    """
    Base class for restrictions relating to `a priori` knowledge of the network.
    """


class InitialDistinctValues(KnowledgeRestriction, ApplicableRestriction):
    """
    The initial node id values are distinct.
    """

    KEY = "unique_value"

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        return len({node.memory.get(cls.KEY, None) for node in network}) == len(network)

    @classmethod
    def apply(cls, network: "NetworkType") -> None:
        for node in network.nodes():
            node.memory[cls.KEY] = node._internal_id

    @classmethod
    def get_help_message(cls, network: "NetworkType") -> str:
        if all(cls.KEY in node.memory for node in network):
            return (
                f"The initial values at `node.memory['{cls.KEY}']` are not distinct for every node.\n"
                + cls.help_message
            )
        else:
            return f"The key '{cls.KEY}' is not present in the memory of every node.\n" + cls.help_message


class NetworkSize(KnowledgeRestriction, ApplicableRestriction):
    """
    The size of the network is known.
    """

    KEY = "network_node_count"

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        network_size = len(network)
        return all(node.memory.get(cls.KEY, -1) == network_size for node in network)

    @classmethod
    def apply(cls, network: "NetworkType") -> None:
        for node in network.nodes():
            node.memory[cls.KEY] = len(network)

    @classmethod
    def get_help_message(cls, network: "NetworkType") -> str:
        if all(cls.KEY in node.memory for node in network):
            return (
                f"The initial values at `node.memory['{cls.KEY}']` are not equal to {len(network)} for every node.\n"
                + cls.help_message
            )
        else:
            return f"The key '{cls.KEY}' is not present in the memory of every node.\n" + cls.help_message



"""
Restrictions related to the communication topology of the underlying graph of the network.

These restrict some communication properties of the network, such as connectivity, shape, and the number of initiators.
"""

from abc import ABC
from typing import TYPE_CHECKING

from networkx import connected_components, is_tree

from pydistsim.message import Message, MetaHeader
from pydistsim.restrictions.base_restriction import CheckableRestriction
from pydistsim.utils.helpers import len_is_not_zero, len_is_one

if TYPE_CHECKING:
    from pydistsim.network.network import NetworkType


class TopologicalRestriction(CheckableRestriction, ABC):
    """
    Restrictions related to the communication topology of the underlying graph of the network.
    """


class Connectivity(TopologicalRestriction):
    """
    The communication topology if strongly connected.
    """

    help_message = "The network is not strongly connected. Consider adding more edges/links to the network."

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        return network.is_connected()


StrongConnectivity = Connectivity
"Type alias to emphasize that strong connectivity is what's checked."


class UniqueInitiator(TopologicalRestriction):
    """
    Only one entity will be able to initiate the algorithm through a spontaneous event.
    """

    help_message = (
        "There is more than one initiator. Consider checking the algorithm initializer method and how many "
        "times it uses an initialization message (`Message(meta_header=NodeAlgorithm.INI)`)."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        def message_is_ini(message: "Message"):
            return message.meta_header == MetaHeader.INITIALIZATION_MESSAGE

        nodes_with_ini = (1 for node in network if len_is_not_zero(filter(message_is_ini, node.inbox)))
        return len_is_one(nodes_with_ini)


class ShapeRestriction(TopologicalRestriction, ABC):
    """
    The communication topology has a specific shape.

    Only the given shape is checked, not the connectivity, even if connectivity is often used with the shape.
    For strong connectivity, use :class:`Connectivity`.
    """


class CompleteGraph(ShapeRestriction):
    """
    The communication topology is a complete graph.

    No self-loops are allowed.
    """

    help_message = (
        "The network topology is not 'complete'. Consider adding more edges/links or using "
        "`NetworkGenerator.generate_complete_network(n)` to generate a new network."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        N = len(network) - 1
        return not any(
            node in neighbors_of_node or len(neighbors_of_node) != N for node, neighbors_of_node in network.adj.items()
        )


class CycleGraph(ShapeRestriction):
    """
    The communication topology is a cycle.

    Strong connectivity is not required as this restriction can be used with :class:`Connectivity`.
    """

    help_message = (
        "The network topology is not a 'cycle'. Consider adding/removing edges or using "
        "`NetworkGenerator.generate_ring_network(n)` to generate a new network."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        network = network.to_undirected()
        return all(len(neighbors_of_node) == 2 for node, neighbors_of_node in network.adj.items()) and len_is_one(
            connected_components(network)
        )


RingGraph = CycleGraph
"Type alias for CycleGraph."


class OrientedCycleGraph(CycleGraph):
    """
    The communication topology is an oriented cycle.

    This means that every node shares the meaning of "left" and "right" with its neighbors.

    *TODO*: This check is not implemented.
    """

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        raise NotImplementedError("ToDo: Implement OrientedCycleGraph restriction.")


OrientedRingGraph = OrientedCycleGraph
"Type alias for OrientedCycleGraph."


class TreeGraph(ShapeRestriction):
    """
    The communication topology is a tree.

    Strong connectivity is not required as this restriction can be used with :class:`Connectivity`.
    """

    help_message = (
        "The network topology is not a 'tree'. Consider adding/removing edges to avoid cycles while also maintaining "
        "connectivity."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        return is_tree(network)


class StarGraph(ShapeRestriction):
    """
    The communication topology is a star.

    A network is a star if there is a node that is connected to all other nodes and the rest of the nodes are only
    connected to the center node.

    If the network has only one or two nodes, it is considered a star.

    Strong connectivity is not required as this restriction can be used with :class:`Connectivity`.
    """

    help_message = (
        "The network topology is not a 'star'. Consider adding/removing edges or using "
        "`NetworkGenerator.generate_star_network(n)` to generate a new network."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        def neig(node):
            return len(set(network.in_neighbors(node)) | set(network.out_neighbors(node)))

        if len(network) <= 2:
            return True

        N = len(network) - 1
        center_count = 0
        others_count = 0
        for node in network.nodes():
            if neig(node) == N:
                center_count += 1
                if center_count > 1:
                    return False
            elif neig(node) == 1:
                others_count += 1
            else:
                return False

        return center_count == 1 and others_count == N


class HyperCubeGraph(ShapeRestriction):
    """
    The communication topology is a hypercube of any dimension.

    Strong connectivity is not required as this restriction can be used with :class:`Connectivity`.

    *TODO*: This check is not implemented.
    """

    help_message = (
        "This restriction is not implemented. Remove it from the list of restrictions and make sure the network is "
        "generated with `NetworkGenerator.generate_hypercube_network(n)`."
    )

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        raise NotImplementedError("ToDo: Implement HyperCubeGraph restriction.\n" + cls.get_help_message(network))


class OrientedHyperCubeGraph(HyperCubeGraph):
    """
    The communication topology is an oriented hypercube.

    Details of the definition can be found in section "3.5 ELECTION IN CUBE NETWORKS" of "Design and Analysis of
    Distributed Algorithms" by Nicola Santoro

    *TODO*: This check is not implemented.
    """

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        raise NotImplementedError("ToDo: Implement OrientedHyperCubeGraph restriction.")



"""Drawing functions for visualizing the simulation."""

import math
from collections.abc import Callable
from enum import StrEnum
from functools import reduce
from typing import TYPE_CHECKING, Union

import matplotlib.pyplot as plt
from matplotlib import animation
from matplotlib.axes import Axes
from matplotlib.collections import PatchCollection
from matplotlib.figure import Figure
from matplotlib.patches import Circle, RegularPolygon
from networkx import draw_networkx_edges as __d_netx_edges
from networkx import draw_networkx_labels as __d_netx_labels

from pydistsim._exceptions import SimulationException
from pydistsim.algorithm.node_algorithm import NodeAlgorithm
from pydistsim.logging import logger
from pydistsim.simulation import Simulation

if TYPE_CHECKING:
    from pydistsim.network.network import NetworkType


class MessageType(StrEnum):
    IN = "Inbox"
    OUT = "Outbox"
    TRANSIT = "Transit"
    LOST = "Lost"


MESSAGE_COLOR = {
    MessageType.IN: "tab:cyan",
    MessageType.OUT: "w",
    MessageType.TRANSIT: "y",
    MessageType.LOST: "r",
}


EDGES_ALPHA = 0.6
NODE_COLORS = [
    "tab:blue",
    "tab:orange",
    "tab:green",
    "tab:red",
    "tab:purple",
    "tab:brown",
    "tab:pink",
    "tab:gray",
    "tab:olive",
    "tab:cyan",
] * 100

MESSAGE_SHAPE_ZORDER = 3
MESSAGE_ANNOTATION_ZORDER = 4


def __get_message_positions_and_orientation(
    source, destination, net: "NetworkType", direction: MessageType
) -> tuple[float, float, float]:
    xd, yd = net.pos[destination]
    xs, ys = net.pos[source]

    angle_in_rads = -math.pi / 2 + math.atan2(yd - ys, xd - xs)

    x = y = None
    if direction == MessageType.OUT:
        offset = 1 / 6
    elif direction == MessageType.IN:
        offset = 7 / 8
    elif direction == MessageType.TRANSIT:
        offset = 1 / 3
    elif direction == MessageType.LOST:
        offset_distance = 10 * (-1 if angle_in_rads < 0 else 1)

        xm = (xs + xd) / 2.0
        ym = (ys + yd) / 2.0

        if xs == xd:  # vertical line
            x = xm + offset_distance
            y = ym
        elif yd == ys:  # horizontal line
            x = xm
            y = ym + offset_distance
        else:  # diagonal line
            slope = (yd - ys) / (xd - xs)
            slope_perpendicular = -1 / slope
            x = xm + offset_distance / (slope_perpendicular**2 + 1) ** 0.5
            y = ym + offset_distance / (slope_perpendicular**2 + 1) ** 0.5 * slope_perpendicular

    if x is None:
        x = xs + (xd - xs) * offset
        y = ys + (yd - ys) * offset

    return x, y, angle_in_rads


def __draw_tree(tree_key: str, net: "NetworkType", axes: Axes):
    """
    Show tree representation of network.

    Attributes:
        tree_key (str):
            key in nodes memory (dictionary) where tree data is stored
            storage format can be a list off tree neighbors or a dict:
                {'parent': parent_node,
                    'children': [child_node1, child_node2 ...]}
    """
    (nodes, edges) = net.get_tree_net(tree_key, return_subnetwork=False)
    if nodes:
        __d_netx_edges(
            net,
            net.pos,
            edges,
            edge_color="tab:brown",
            width=2.5,
            alpha=EDGES_ALPHA + 0.2,
            ax=axes,
        )


def __draw_nodes(net, axes, node_colors={}, node_radius={}, radius_default=8.0):
    if isinstance(node_colors, str):
        node_colors = {node: node_colors for node in net.nodes()}
    nodeCircles = []
    for n in net.nodes():
        c = Circle(
            tuple(net.pos[n]),
            node_radius.get(n, radius_default),
            color=node_colors.get(n, "r"),
            ec="k",
            lw=1.0,
            ls="solid",
            picker=3,
        )
        nodeCircles.append(c)
    node_collection = PatchCollection(nodeCircles, match_original=True)
    node_collection.set_picker(3)
    axes.add_collection(node_collection)
    return node_collection


def __create_and_get_color_labels(net, algorithm=None, subclusters=None, figure: Figure = None, show_messages=True):
    proxy_kwargs = {
        "xy": (0, 0),
        "radius": 8.0,
        "ec": "k",
        "lw": 1.0,
        "ls": "solid",
    }

    node_colors = {}
    if algorithm:
        color_map = {}
        if isinstance(algorithm, NodeAlgorithm):
            for ind, status in enumerate(algorithm.Status.__members__):
                color_map.update({status: NODE_COLORS[ind]})
            if figure:
                # Node status legend
                proxy = []
                labels = []
                for status, color in list(color_map.items()):
                    proxy.append(
                        Circle(
                            color=color,
                            **proxy_kwargs,
                        )
                    )
                    labels.append(status)
                figure.legends = []
                figure.legend(
                    proxy,
                    labels,
                    loc="center right",
                    ncol=1,
                    bbox_to_anchor=(1, 0.8),
                    title="Statuses for %s:" % algorithm.name,
                )
                if show_messages:
                    # Message legend
                    figure.legend(
                        [
                            Circle(
                                color=MESSAGE_COLOR[msg],
                                **proxy_kwargs,
                            )
                            for msg in (
                                MessageType.IN,
                                MessageType.OUT,
                                MessageType.TRANSIT,
                                MessageType.LOST,
                            )
                        ],
                        ["Inbox", "Outbox", "Transit", "Lost"],
                        loc="center right",
                        ncol=1,
                        bbox_to_anchor=(1, 0.1),
                        title="Messages:",
                    )
                plt.subplots_adjust(left=0.1, bottom=0.1, right=0.99)

        for n in net.nodes():
            if n.status == "" or n.status not in list(color_map.keys()):
                node_colors[n] = "k"
            else:
                node_colors[n] = color_map[n.status]
    elif subclusters:
        for i, sc in enumerate(subclusters):
            for n in sc:
                if n in node_colors:
                    node_colors[n] = "k"
                else:
                    node_colors[n] = NODE_COLORS[i]
    return node_colors


def __draw_edges(net, edges, axes):
    __d_netx_edges(net, net.pos, alpha=EDGES_ALPHA, edgelist=edges, ax=axes)


def __draw_messages(net: "NetworkType", axes: Axes, message_radius: float):
    MESSAGE_LINE_WIDTH = 1.0
    patch_kwargs = {
        "numVertices": 3,
        "radius": message_radius,
        "lw": MESSAGE_LINE_WIDTH,
        "ls": "solid",
        "picker": 3,
        "zorder": MESSAGE_SHAPE_ZORDER,
        "ec": "k",
    }

    msg_artists = []

    message_collection = {
        node: {
            MessageType.OUT: [
                ([msg.destination] if msg.destination is not None else list(net.adj[node].keys()))
                for msg in node.outbox
            ],
            MessageType.IN: [[msg.source] for msg in node.inbox],
            MessageType.TRANSIT: [
                [other_node for msg in net.get_transit_messages(node, other_node) if msg.source == node]
                for other_node in net.out_neighbors(node)
                if net.get_transit_messages(node, other_node)
            ],
            MessageType.LOST: [
                [other_node for msg in net.get_lost_messages(node, other_node) if msg.source == node]
                for other_node in net.out_neighbors(node)
                if net.get_lost_messages(node, other_node)
            ],
        }
        for node in net.nodes()
    }

    for node in message_collection:
        messages_type_dict = message_collection[node]

        msg_dict = {}
        for msg_type in messages_type_dict:
            dest_lists = messages_type_dict[msg_type]

            msg_dict[msg_type] = {}
            for dest in reduce(lambda x, y: x + y, dest_lists, []):
                if dest is None:
                    continue

                src = node if msg_type != MessageType.IN else dest
                dst = dest if msg_type != MessageType.IN else node

                count = msg_dict[msg_type].get((src, dst), 0)
                msg_dict[msg_type][(src, dst)] = count + 1

        for msg_type in msg_dict:
            for (src, dst), count in msg_dict[msg_type].items():
                if not src or not dst:
                    continue  # Defensive check

                x, y, rads_orientation = __get_message_positions_and_orientation(src, dst, net, msg_type)
                triangle_artist = RegularPolygon(
                    (x, y),
                    orientation=rads_orientation,
                    **patch_kwargs,
                    fc=MESSAGE_COLOR[msg_type],
                    label=msg_type,
                )
                axes.annotate(
                    f"{count}",
                    (x + 5, y + 5),
                    color="k",
                    fontsize=8,
                    zorder=MESSAGE_ANNOTATION_ZORDER,
                )
                msg_artists.append(triangle_artist)

    if msg_artists:
        message_collection = PatchCollection(msg_artists, match_original=True)
        message_collection.set_picker(3)
        axes.add_collection(message_collection)


def __draw_labels(net: "NetworkType", node_size, dpi):
    label_pos = {}
    from math import sqrt

    label_delta = 1.5 * sqrt(node_size) * dpi / 100
    for n in net.nodes():
        label_pos[n] = net.pos[n].copy() + label_delta
    __d_netx_labels(
        net,
        label_pos,
        labels=net.labels,
        horizontalalignment="left",
        verticalalignment="bottom",
    )


def draw_current_state(
    sim: Union["Simulation", "NetworkType"],
    axes: Axes = None,
    clear: bool = True,
    tree_key: str = None,
    dpi: int = 100,
    node_radius: int = 10,
    node_positions: dict | Callable = None,
    node_colors: dict | Callable = None,
    edge_filter: list = None,
    show_messages: bool = True,
    show_labels: bool = True,
    node_labels: dict | Callable = None,
):
    """
    Function to draw the current state of the simulation or network. This function is used to visualize the network
    and the messages in the network.
    Automatically determines the current algorithm and draws the network accordingly. This includes a mapping of
    node colors to the status of the nodes, as well as the messages in the network.

    :param sim: Simulation or Network object
    :param axes: matplotlib axes object
    :param clear: boolean to clear the axes before drawing
    :param tree_key: key in nodes memory (dictionary) where tree data is stored
    :param dpi: dots per inch
    :param node_radius: radius of nodes
    :param node_positions: dictionary of node positions or function to get node positions
    :param node_colors: dictionary of node colors or function to get node colors
    :param edge_filter: list of edges to draw
    :param show_messages: boolean to show messages in the network
    :param show_labels: boolean to show labels of nodes
    :param node_labels: dictionary of node labels or function to get node labels
    :return: matplotlib figure object
    """

    if isinstance(sim, Simulation):
        net = sim.network

        try:
            currentAlgorithm = sim.get_current_algorithm()
        except SimulationException:
            currentAlgorithm = None
    else:
        net = sim
        currentAlgorithm = None

    if node_positions:
        pos_aux = net.pos
        net.pos = node_positions() if callable(node_positions) else node_positions

    if node_labels:
        aux_labels = net.labels
        net.labels = node_labels() if callable(node_labels) else node_labels

    if axes is None:
        with plt.ioff():
            fig = plt.figure()
            axes = fig.add_subplot(1, 1, 1)

    if clear:
        axes.clear()

    axes.imshow(net.environment.image, vmin=0, cmap="binary_r", origin="lower")

    if tree_key:
        __draw_tree(tree_key, net, axes)

    __draw_edges(net, edge_filter, axes)

    if not node_colors:
        node_colors = __create_and_get_color_labels(
            net, algorithm=currentAlgorithm, figure=axes.figure, show_messages=show_messages
        )
    elif callable(node_colors):
        node_colors = node_colors()

    __draw_nodes(net, axes, node_colors, radius_default=node_radius)
    if show_messages:
        __draw_messages(net, axes, message_radius=3 * node_radius / 4)

    if show_labels:
        __draw_labels(net, node_radius, dpi)

    step_text = " (step %d)" % sim.algorithmState["step"] if isinstance(currentAlgorithm, NodeAlgorithm) else ""
    axes.set_title((currentAlgorithm.name if currentAlgorithm else "") + step_text)

    axes.axis("off")
    # remove as much whitespace as possible
    axes.figure.tight_layout()
    axes.figure.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=None)

    if node_positions:
        net.pos = pos_aux

    if node_labels:
        net.labels = aux_labels

    return axes.figure


def create_animation(
    sim: "Simulation",
    figsize=None,
    dpi: int = 100,
    milliseconds_per_frame: int = 300,
    frame_limit: int = 2000,
    **kwargs,
) -> animation.FuncAnimation:
    """
    Create an animation of the simulation.

    Example for visualizing in Jupyter Notebook:

    .. code-block:: python

        anim = create_animation(sim)

        video = anim.to_html5_video()

        from IPython.display import HTML
        HTML(video)

    Example for saving as a video file:

    .. code-block:: python

        from matplotlib.animation import FFMpegFileWriter

        moviewriter = FFMpegFileWriter()
        anim = draw.create_animation(sim)

        anim.save("flood.mp4", writer=moviewriter)

    :param sim: Simulation object
    :param figsize: figure size
    :param dpi: dots per inch
    :param milliseconds_per_frame: milliseconds per frame
    :param frame_limit: limit of frames, default is 2000
    :param kwargs: additional keyword arguments to pass to the :func:`draw_current_state` function
    :return: animation object
    """

    with plt.ioff():  # Turn off interactive mode
        fig = plt.figure(figsize=figsize, dpi=dpi)
        ax = fig.add_subplot(111)

        def draw_frame(frame_index):
            if frame_index == 0:
                sim.reset()

            if not sim.is_halted() or frame_index == 0:
                draw_current_state(sim, ax, dpi=dpi, **kwargs)

            sim.run(1)

        def frame_count():
            frame_index = 0

            def should_continue():

                if frame_limit and frame_index >= frame_limit:
                    logger.warning("Frame limit reached.")
                    return False

                if not sim.is_halted():
                    logger.debug(f"Frame {frame_index}, simulation still running.")
                    return True

                if frame_index == 0:
                    logger.debug(f"Frame {frame_index}, simulation not started.")
                    return True

                return False

            while True:
                if should_continue():
                    yield frame_index
                    frame_index += 1
                else:
                    break

        return animation.FuncAnimation(
            fig,
            func=draw_frame,
            frames=frame_count,
            interval=milliseconds_per_frame,
            blit=False,
            cache_frame_data=False,
        )



from pydistsim.algorithm.base_algorithm import BaseAlgorithm


class NetworkAlgorithm(BaseAlgorithm):
    """
    Abstract base class for specific centralized algorithms.

    This class is used as base class holding real network algorithm
    classes in its __subclassess__ for easy instantiation

    Method __init__ and run should be implemented in subclass.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_halted = False

    def step(self, check_restrictions: bool, step: int):
        if check_restrictions:
            self.check_restrictions()

        self._is_halted = True
        return self.run()

    def run(self):
        raise NotImplementedError

    def is_halted(self):
        return self._is_halted



from copy import copy
from functools import reduce

from pydistsim.utils.memory import MemoryStructure


class Positions(MemoryStructure):
    """Class to represent subclusters positions data in node memory.

    Position instances initializer has only one attribute ``subclusters`` a
    list of dictionaries in form {node: position_array,}. Node can be part of
    multiple subclusters.

    """

    def __init__(self, subclusters=None):
        self.subclusters = []
        if subclusters is not None:
            # list of dictionaries of positioned subclusters
            assert isinstance(subclusters, list)
            # create subclusters by copying from given list and its dicts
            ini_subclusters = list(subclusters)
            self.subclusters = []
            for subcluster in ini_subclusters:
                new_subcluster = {}
                for n, p in list(subcluster.items()):
                    new_subcluster.update({n: copy(p)})
                self.subclusters.append(new_subcluster)
        self.old_style_positions = {}
        self.old_style_subclusters = []

    @classmethod
    def create(cls, obj):
        """Create Positions instance from obj which can be dict or list.

        If it's dict this method wraps it in list. Returns Position object
        dereferenced from the initial list or dict.

        For example ``obj`` can be ``net.pos`` dict.

        """
        if not isinstance(obj, cls):
            if not isinstance(obj, list):
                assert isinstance(obj, dict)
                obj = [obj]
            obj = cls(obj)
        return obj

    @property
    def subclusters_nodes(self):
        """Returns list of subclusters without positions."""
        cl = []
        for c in self.subclusters:
            cl.append(list(c.keys()))
        return cl

    def set_pos_copy(self, positions):
        """Empty all subclusters and copy in data from ``positions``."""
        assert isinstance(positions, Positions)
        self.subclusters = []
        for subcluster in positions.subclusters:
            new_subcluster = {}
            for n, p in list(subcluster.items()):
                new_subcluster.update({n: copy(p)})
            self.subclusters.append(new_subcluster)

    def __len__(self):
        return len(self.subclusters)

    def get_nodes(self):
        """Returns nodes in all subclusters."""
        nodes = []
        for subcluster in self.subclusters:
            nodes += [node for node in subcluster if node not in nodes]
        return nodes

    def get_largest_subcluster(self):
        """Returns new Positions instance with only one largest subcluster.
        If multiple subclusters have maximum number of nodes, first one is
        returned."""
        return Positions([reduce(lambda x, y: len(x) > len(y) and x or y, self.subclusters)])

    def get_dic(self):
        dic = {}
        for i, sc in enumerate(self.subclusters):
            dic["subclusters[%d] (%d nodes)" % (i, len(sc))] = sc
        return dic

    def __str__(self):
        return str(self.subclusters)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.subclusters[key]
        return self.__getattribute__(key)



import os  # @Reimport
import sys
from datetime import datetime

import numpy
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.collections import LineCollection
from matplotlib.figure import Figure
from matplotlib.patches import Circle
from networkx.drawing.nx_pylab import draw_networkx_edges
from PySide6.QtCore import SIGNAL, QEvent, QRect, QSize, QThread
from PySide6.QtGui import QCursor
from PySide6.QtWidgets import QFileDialog, QMainWindow, QMenu, QMessageBox

from pydistsim import Simulation
from pydistsim.algorithm import NodeAlgorithm
from pydistsim.algorithm.base_algorithm import BaseAlgorithm
from pydistsim.observers import AlgorithmObserver, SimulationObserver
from pydistsim.utils.npickle import read_pickle, write_pickle

try:
    from .dictionarytreemodel import DictionaryTreeModel
    from .simulationui import Ui_SimulationWindow
except ImportError:
    from pydistsim.gui.dictionarytreemodel import DictionaryTreeModel
    from pydistsim.gui.simulationui import Ui_SimulationWindow

from copy import deepcopy

from pydistsim import logger
from pydistsim.gui.drawing import draw_current_state
from pydistsim.utils.localization.helpers import align_clusters, get_rms
from pydistsim.utils.memory.positions import Positions


class QThreadObserver(AlgorithmObserver, SimulationObserver):
    def __init__(self, q_thread: QThread, *args, **kwargs) -> None:
        self.q_thread = q_thread
        super().__init__(*args, **kwargs)

    def on_step_done(self, algorithm: "BaseAlgorithm") -> None:
        self.q_thread.emit(
            SIGNAL("updateLog(QString)"),
            "[{}] Step {} finished",
            algorithm.name,
            algorithm.simulation.algorithmState["step"],
        )

    def on_state_changed(self, simulation: Simulation) -> None:
        self.q_thread.emit(SIGNAL("redraw()"))

    def on_algorithm_finished(self, algorithm: "BaseAlgorithm") -> None:
        self.q_thread.emit(SIGNAL("updateLog(QString)"), "[%s] Algorithm finished" % (algorithm.name))

    def on_network_changed(self, simulation: Simulation) -> None:
        self.q_thread.emit(SIGNAL("updateLog(QString)"), "Network loaded")
        self.q_thread.emit(SIGNAL("redraw()"))


class SimulationThread(Simulation, QThread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.add_observers(QThreadObserver(self))

    def __del__(self):
        self.exiting = True
        self.wait()


class SimulationGui(QMainWindow):
    def __init__(self, net=None, parent=None, fname=None):
        super().__init__()

        self.ui = Ui_SimulationWindow()
        self.ui.setupUi(self)

        if fname:
            self.set_title(fname)

        # context menu
        self.ui.nodeInspector.addAction(self.ui.actionCopyInspectorData)
        self.ui.nodeInspector.addAction(self.ui.actionShowLocalizedSubclusters)
        # callbacks
        self.ui.actionCopyInspectorData.triggered.connect(self.on_actionCopyInspectorData_triggered)
        self.ui.actionShowLocalizedSubclusters.triggered.connect(self.on_actionShowLocalizedSubclusters_triggered)

        self.dpi = 72
        # take size of networDisplayWidget
        self.fig = Figure((700 / self.dpi, 731 / self.dpi), self.dpi, facecolor="0.9")
        self.canvas = FigureCanvas(self.fig)
        self.canvas.setParent(self.ui.networkDisplayWidget)
        self.nav = NavigationToolbar(self.canvas, self.ui.networkDisplayWidget, coordinates=True)
        self.nav.setGeometry(QRect(0, 0, 651, 36))
        self.nav.setIconSize(QSize(24, 24))

        self.axes = self.fig.add_subplot(111)
        # matplotlib.org/api/figure_api.html#matplotlib.figure.SubplotParams
        self.fig.subplots_adjust(left=0.03, right=0.99, top=0.92)

        if net:
            self.init_sim(net)

        self.connect(self.ui.showNodes, SIGNAL("stateChanged(int)"), self.refresh_visibility)
        self.connect(self.ui.showEdges, SIGNAL("stateChanged(int)"), self.refresh_visibility)
        self.connect(self.ui.showMessages, SIGNAL("stateChanged(int)"), self.refresh_visibility)
        self.connect(self.ui.showLabels, SIGNAL("stateChanged(int)"), self.refresh_visibility)
        self.connect(self.ui.redrawNetworkButton, SIGNAL("clicked(bool)"), self.redraw)
        self.connect(self.ui.treeGroupBox, SIGNAL("toggled(bool)"), self.refresh_visibility)
        self.connect(self.ui.tree_key, SIGNAL("textEdited(QString)"), self.redraw)
        self.connect(self.ui.propagationError, SIGNAL("toggled(bool)"), self.refresh_visibility)
        self.connect(self.ui.locKey, SIGNAL("textEdited(QString)"), self.redraw)
        # callbacks
        self.ui.actionOpenNetwork.triggered.connect(self.on_actionOpenNetwork_triggered)
        self.ui.actionSaveNetwork.triggered.connect(self.on_actionSaveNetwork_triggered)
        self.ui.actionRun.triggered.connect(self.on_actionRun_triggered)
        self.ui.actionStep.triggered.connect(self.on_actionStep_triggered)
        self.ui.actionReset.triggered.connect(self.on_actionReset_triggered)

        self.canvas.mpl_connect("pick_event", self.on_pick)

    def handleInspectorMenu(self, pos):
        menu = QMenu()
        menu.addAction("Add")
        menu.addAction("Delete")
        menu.exec_(QCursor.pos())

    def init_sim(self, net):
        self.net = net
        self.sim = Simulation(net)
        self.connect(self.sim, SIGNAL("redraw()"), self.redraw)
        self.connect(self.sim, SIGNAL("updateLog(QString)"), self.update_log)
        self.redraw()

    def update_log(self, text):
        """Add item to list widget"""
        logger.debug("Added item to list widget: " + text)
        self.ui.logListWidget.insertItem(0, text)
        # self.ui.logListWidget.sortItems()

    def redraw(self):
        self.refresh_network_inspector()
        self.draw_network()
        self.reset_zoom()
        self.refresh_visibility()

    def draw_network(self):
        draw_current_state(
            self.sim, self.axes, clear=True, tree_key=self.ui.tree_key.text(), locKey=self.ui.locKey.text()
        )

    def refresh_visibility(self):
        try:
            self.node_collection.set_visible(self.ui.showNodes.isChecked())
            self.edge_collection.set_visible(self.ui.showEdges.isChecked())
            for label in list(self.label_collection.values()):
                label.set_visible(self.ui.showLabels.isChecked())
            self.tree_collection.set_visible(self.ui.treeGroupBox.isChecked())
            self.ini_error_collection.set_visible(self.ui.propagationError.isChecked())
            self.propagation_error_collection.set_visible(self.ui.propagationError.isChecked())
            # sould be last, sometimes there are no messages
            self.message_collection.set_visible(self.ui.showMessages.isChecked())
        except AttributeError:
            logger.warning("Refresh visibility warning.")
        self.canvas.draw()

    def reset_zoom(self):
        self.axes.set_xlim((0, self.net.environment.image.shape[1]))
        self.axes.set_ylim((0, self.net.environment.image.shape[0]))

    def set_title(self, fname):
        new = " - ".join([str(self.windowTitle()).split(" - ")[0], str(fname)])
        self.setWindowTitle(new)

    def refresh_network_inspector(self):
        niModel = DictionaryTreeModel(dic=self.net.get_dic())
        self.ui.networkInspector.setModel(niModel)
        self.ui.networkInspector.expandToDepth(0)

    """
    Callbacks
    """

    def on_actionRun_triggered(self):
        self.ui.logListWidget.clear()
        logger.debug("running on_actionRun_triggered")
        self.sim.stepping = True
        self.sim.run()

    def on_actionStep_triggered(self):
        logger.debug("running on_actionStep_triggered")
        self.sim.run(self.ui.stepSize.value())

    def on_actionReset_triggered(self):
        logger.debug("running on_actionReset_triggered")
        self.sim.reset()
        self.redraw()

    def on_actionCopyInspectorData_triggered(self):
        string = "Node inspector data\n-------------------"
        # raise()
        for qModelIndex in self.ui.nodeInspector.selectedIndexes():
            string += "\n" + qModelIndex.internalPointer().toString("    ")

        clipboard = app.clipboard()
        clipboard.setText(string)
        event = QEvent(QEvent.Clipboard)
        app.sendEvent(clipboard, event)

    def on_actionShowLocalizedSubclusters_triggered(self):
        if len(self.ui.nodeInspector.selectedIndexes()) == 1:
            qModelIndex = self.ui.nodeInspector.selectedIndexes()[0]
            treeItem = qModelIndex.internalPointer()
            assert isinstance(treeItem.itemDataValue, Positions)

            estimated = deepcopy(treeItem.itemDataValue)
            estimatedsub = estimated.subclusters[0]
            # rotate, translate and optionally scale
            # w.r.t. original positions (pos)
            align_clusters(Positions.create(self.net.pos), estimated, True)
            net = self.net.subnetwork(list(estimatedsub.keys()), pos=estimatedsub)

            self.draw_network(net=net, drawMessages=False)

            edge_pos = numpy.asarray([(self.net.pos[node], estimatedsub[node][:2]) for node in net])
            error_collection = LineCollection(edge_pos, colors="r")
            self.axes.add_collection(error_collection)

            rms = get_rms(self.net.pos, estimated, scale=False)
            self.update_log("rms = %.3f" % rms)
            self.update_log(
                "localized = %.2f%% (%d/%d)"
                % (
                    len(estimatedsub) * 1.0 / len(self.net.pos) * 100,
                    len(estimatedsub),
                    len(self.net.pos),
                )
            )

    def on_actionSaveNetwork_triggered(self, *args):
        default_filetype = "gz"
        start = datetime.now().strftime("%Y%m%d") + default_filetype

        filters = ["Network pickle (*.gz)", "All files (*)"]
        selectedFilter = "Network pickle (gz)"
        filters = ";;".join(filters)

        fname = QFileDialog.getSaveFileName(self, "Choose a filename", start, filters, selectedFilter)[0]
        if fname:
            try:
                write_pickle(self.net, fname)
            except Exception as e:
                QMessageBox.critical(
                    self,
                    "Error saving file",
                    str(e),
                    QMessageBox.Ok,
                    QMessageBox.NoButton,
                )
            else:
                self.set_title(fname)

    def on_actionOpenNetwork_triggered(self, *args):
        default_filetype = "gz"
        start = datetime.now().strftime("%Y%m%d") + default_filetype

        filters = ["Network pickle (*.gz)", "All files (*)"]
        selectedFilter = "Network pickle (gz)"
        filters = ";;".join(filters)

        fname = QFileDialog.getOpenFileName(self, "Choose a file to open", start, filters, selectedFilter)[0]
        if fname:
            try:
                logger.debug("opening " + fname)
                net = read_pickle(fname)
                self.init_sim(net)
            except Exception as e:
                logger.exception("Error opening file %s" % str(e))
                QMessageBox.critical(
                    self,
                    "Error opening file",
                    str(e),
                    QMessageBox.Ok,
                    QMessageBox.NoButton,
                )
            else:
                self.set_title(fname)

    def on_pick(self, event):
        if (
            event.artist == self.node_collection
            or event.artist == self.propagation_error_collection
            or event.artist == self.ini_error_collection
        ):
            for ind in event.ind:
                self.on_pick_node(self.drawnNet.nodes()[ind])
        elif event.artist == self.message_collection:
            for ind in event.ind:
                self.on_pick_message(self.messages[ind])
        self.canvas.draw()

    def on_pick_node(self, node):
        niModel = DictionaryTreeModel(dic=node.get_dic())
        # TODO: self.ui.nodeInspectorLabel.setText('Node inspector: node %d' % node._internal_id)
        self.ui.nodeInspector.setModel(niModel)
        self.ui.nodeInspector.expandToDepth(0)

    def on_pick_message(self, message):
        self.ui.logListWidget.insertItem(0, "Pick message %s " % repr(message))


from IPython.lib.guisupport import get_app_qt4, start_event_loop_qt4


def create_window(window_class, **kwargs):
    """Create a QT window in Python, or interactively in IPython with QT GUI
    event loop integration.
    """
    global app

    app = get_app_qt4(sys.argv)
    app.references = set()

    net = None
    fname = None
    if len(sys.argv) > 1:
        fname = sys.argv[1]
        if os.path.exists(fname):
            net = read_pickle(fname)
        else:
            QMessageBox.critical(
                None,
                "Error opening file %s",
                fname,
                QMessageBox.Ok,
                QMessageBox.NoButton,
            )

    window = window_class(net, fname)
    app.references.add(window)
    window.show()

    start_event_loop_qt4(app)
    return window


def main():
    global simgui
    simgui = create_window(SimulationGui)



from numpy import arctan2, array, concatenate, dot, isnan, mod, nan, outer, pi, sqrt
from numpy.lib._type_check_impl import imag, real
from numpy.linalg import det, eig

from pydistsim.utils.localization.stitchsubclusterselectors import (
    MaxCommonNodeSelector,
    StitchSubclusterSelectorBase,
)


class BaseStitcher:
    """
    Base class for stitching two clusters.

    Subclasses should implement own methods for subcluster stitching named:
        stitch_subclusters
    These methods should take two subclusters as dictionary {node: position}
    where position is array([x,y,theta]) with at least one common node and
    return rotation matrix R, scaling factor s and translation vector t.

    If stitch_subclusters method can't be sure that all parameters are correct
    i.e. when common nodes are not in generic formation then it should return
    None for all parameters.

    """

    def __init__(self, selector=None):
        self.selector = selector or MaxCommonNodeSelector()
        assert isinstance(self.selector, StitchSubclusterSelectorBase)

    def stitch(self, dst, src, do_intra=True):
        """
        Stitch src cluster to dst cluster based on selector and its criteria.

        1. Stitch all src subclusters to dst.
        2. Append unstitched source subclusters to dst cluster.
        3. If do_intra stitch subclusters internally.

        src and dst clusters are in format defined in
        pydistsim.utils.memory.positions.Positions.subclusters:
        [{node1: array(x1,y1), node2: array(x2,y2), ...}, ...]

        """

        ###### externalStitch ######
        stitched = self._stitch(dst, src)

        # First remove incomplete stitches from stitched
        for k, v in list(stitched.items()):
            if any(x is None for x in v):
                stitched.pop(k)
        # Then using stitched dictionary keys find out which subclusters are
        # not stitched and append them to dst.subclusters
        src_stitched = [s[1] for s in list(stitched.keys())]
        for src_sc_index in range(len(src)):
            if src_sc_index not in src_stitched:
                new_subcluster = {}
                for node, pos in list(src[src_sc_index].items()):
                    new_subcluster[node] = pos
                dst.subclusters.append(new_subcluster)

        if do_intra:
            self.intrastitch(dst)

    def intrastitch(self, dst):
        self._stitch(dst, dst, is_intra=True)

    def _stitch(self, dst, src, is_intra=False):
        """
        Iteratively:
            * selects subclusters to stitch using self.selector
            * stitch - find out R, s and t
            * append transformed node positions from src to dst subcluster
        Return stitched dictionary where:
            keys - are pairs of indexes of all stitched subclusters
            values - are their R, s and t transformation parameters
                     or all of them are None if can't be calculated reliably
                     or () if src subcluster is subset of dst

        """
        stitched = {}
        while True:
            dstSubIndex, srcSubIndex = self.selector.select(dst, src, stitched, is_intra)

            if dstSubIndex is None and srcSubIndex is None:
                break

            # stitch srcSub to dstSub using given method
            R, s, t = self.stitch_subclusters(dst[dstSubIndex], src[srcSubIndex])
            if any(x is None for x in (R, s, t)):  # skip unreliable stitches
                stitched[(dstSubIndex, srcSubIndex)] = (R, s, t)
                stitched[(srcSubIndex, dstSubIndex)] = (R, s, t)
                continue

            # merge subclusters: append src nodes in dst
            # TODO: apply flip ambiguity condition for new subcluster
            for node in list(src[srcSubIndex].keys()):
                if node not in dst[dstSubIndex]:  # append only new
                    pos = src[srcSubIndex][node][:2]
                    try:
                        ori = src[srcSubIndex][node][2]
                    except IndexError:
                        ori = nan
                    dst[dstSubIndex][node] = self.transform(R, s, t, pos, ori)
                    assert not isnan(dst[dstSubIndex][node][:2]).any()

            stitched[(dstSubIndex, srcSubIndex)] = (R, s, t)

            if is_intra:
                dst.subclusters.pop(srcSubIndex)  # remove stitched subclusters
                stitched = {}  # reset stitched since pop has changed indexes

        return stitched

    def align(self, dst, src):
        """Align (modify) src w.r.t. dst."""

        for srcSubIndex in range(len(src)):
            R, s, t = self.stitch_subclusters(dst[0], src[srcSubIndex])
            for node in list(src[srcSubIndex].keys()):
                src_pos = src[srcSubIndex][node][:2]
                src[srcSubIndex][node] = self.transform(R, s, t, src_pos)

    def transform(self, R, s, t, pos, ori=nan):
        """Transform node position."""
        assert all(x is not None for x in (R, s, t))
        assert not imag(R).any()
        R = real(R)
        if not isnan(ori):
            # angle of rotation matrix R in ccw
            Rtheta = arctan2(R[1, 0], R[0, 0])
            ori = mod(ori - Rtheta, 2 * pi)
        return concatenate((t + dot(dot(s, R), pos), [ori]))

    # ------------------ generic helper methods -------------------------------

    def _get_common_nodes(self, dstSubPos, srcSubPos):
        return list(set(dstSubPos.keys()) & set(srcSubPos.keys()))

    def _get_centroids(self, commonNodes, dstSubPos, srcSubPos, w_d=None):
        if w_d is None:
            w_d = {node: 1.0 / len(commonNodes) for node in commonNodes}
        p_s = array([0.0, 0.0])
        p_d = array([0.0, 0.0])
        for cn in commonNodes:
            p_s += srcSubPos[cn][:2] * w_d[cn]
            p_d += dstSubPos[cn][:2] * w_d[cn]
        return (p_s, p_d, w_d)

    def _get_rotation_matrix_horn(self, commonNodes, dstSubPos, srcSubPos, p_d, p_s):
        """Horn method for calculating rotation matrix."""
        M = sum([outer((dstSubPos[cn][:2] - p_d), (srcSubPos[cn][:2] - p_s)) for cn in commonNodes])
        D, V = eig(dot(M.T, M))
        R = dot(
            M,
            (1 / sqrt(D[0]) * outer(V[:, 0], V[:, 0].conj().T) + 1 / sqrt(D[1]) * outer(V[:, 1], V[:, 1].conj().T)),
        )
        # only rotation, if det(M)<0 it is reflection (Horn et.al)
        if det(M) < 0:
            # TODO: Umeyama
            assert det(R) < 0  # Umeyama1991
        assert not isnan(R).any()
        return R



from pydistsim.benchmark import MetricCollector
from pydistsim.demo_algorithms.broadcast import Flood
from pydistsim.message import Message
from pydistsim.network import NetworkGenerator
from pydistsim.simulation import Simulation
from pydistsim.utils.testing import PyDistSimTestCase


class CustomMetricCollector(MetricCollector):
    """
    A test observer that raises an exception when notified.
    """

    events = ["example_custom_event"]

    def __init__(self) -> None:
        super().__init__()
        self.example_custom_event_msgs = []

    def on_example_custom_event(self, message: "Message"):
        self.example_custom_event_msgs.append(message.source)

    def create_report(self):
        report = super().create_report()
        report["example_custom_event_msgs_sources"] = self.example_custom_event_msgs
        return report


class CustomAlgorithm(Flood):

    @Flood.Status.IDLE
    def receiving(self, node, message):
        super().receiving_IDLE(node, message)
        self.notify_observers("example_custom_event", message)


class TestMetricCollector(PyDistSimTestCase):
    def setUp(self):
        self.observer = MetricCollector()
        self.net = NetworkGenerator(10, directed=False).generate_random_network()
        self.sim = Simulation(self.net)
        self.sim.algorithms = ((Flood, {"informationKey": "greet"}),)

        self.initiator = self.net.nodes_sorted()[0]
        self.initiator.memory["greet"] = "HELLO"

    def test_all(self):
        self.sim.add_observers(self.observer)
        self.sim.run(100_000)
        report = self.observer.create_report()

        assert "Qty. of messages sent" in report and report["Qty. of messages sent"] > 0
        assert "Qty. of messages delivered" in report and report["Qty. of messages delivered"] > 0
        assert "Qty. of status changes" in report and report["Qty. of status changes"] > 0
        assert "Qty. of steps" in report and report["Qty. of steps"] > 0


class TestWrongEvent(PyDistSimTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.observer = MetricCollector()
        cls.net = NetworkGenerator(10, directed=False).generate_random_network()

    def test_wrong_event(self):
        # Nothing should happen
        self.net.add_observers(self.observer)
        self.net.notify_observers("test", None)


class TestCustomMetricCollector(PyDistSimTestCase):
    def setUp(self):
        self.observer = CustomMetricCollector()
        self.net = NetworkGenerator(10, directed=False).generate_random_network()
        self.sim = Simulation(self.net)
        self.sim.algorithms = ((CustomAlgorithm, {"informationKey": "greet"}),)

        self.initiator = self.net.nodes_sorted()[0]
        self.initiator.memory["greet"] = "HELLO"

    def test_all(self):
        self.sim.add_observers(self.observer)
        self.sim.run(100_000)
        report = self.observer.create_report()

        assert "Qty. of messages sent" in report and report["Qty. of messages sent"] > 0
        assert "Qty. of messages delivered" in report and report["Qty. of messages delivered"] > 0
        assert "Qty. of status changes" in report and report["Qty. of status changes"] > 0
        assert "Qty. of steps" in report and report["Qty. of steps"] > 0
        assert "example_custom_event_msgs_sources" in report and len(report["example_custom_event_msgs_sources"]) > 0



"""
Restrictions related to reliability, faults, and their detection.
"""

from abc import ABC
from enum import StrEnum
from typing import TYPE_CHECKING

from pydistsim.logging import logger
from pydistsim.restrictions.base_restriction import CheckableRestriction

if TYPE_CHECKING:
    from pydistsim.network.network import NetworkType


class NotImplementedTypes(StrEnum):
    edge_failure = "edge failure and its detection"
    entity_failure = "entity failure"
    message_corruption = "message corruption"


def __log_failure_not_implemented__(cls, failure_type: str) -> None:
    logger.debug(
        f"{cls.__name__} is not fully implemented as {failure_type} is not supported. This restriction only "
        "checks for message loss."
    )


class ReliabilityRestriction(CheckableRestriction, ABC):
    """
    Base class for restrictions related to reliability.
    """

    help_message = (
        "Message loss is enabled. Choose a value for `network.behavioral_properties` such that `message_loss_indicator`"
        " is `None`.\nThe property NetworkBehaviorModel.RandomDelayCommunication should be a good example."
    )


class EdgeFailureDetection(ReliabilityRestriction):
    """
    Whether or not all entities in the network are able to detect a fail in one of its edges and, following the
    failure, detect if it was reactivated.

    In this context, a failure is a temporary loss of the ability to send messages over an edge. Not to be confused with
    delay or a message loss, even though a failure might cause these effects.

    *TODO*: Edge failures are not implemented by the simulation.
    """

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        __log_failure_not_implemented__(cls, NotImplementedTypes.edge_failure)
        raise NotImplementedError("ToDo: implement dynamic edge failure and optional detection")


class EntityFailureDetection(ReliabilityRestriction):
    """
    For all nodes x in the network,  all in- and out-neighbors of x can detect whether x has failed and, following its
    failure, whether it has recovered.

    In this context, a failure is a temporary loss of the ability of a node to react to events.

    *TODO*: Entity failures are not implemented by the simulation.
    """

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        __log_failure_not_implemented__(cls, NotImplementedTypes.entity_failure)
        raise NotImplementedError("ToDo: implement dynamic entity failure and optional detection")


class TotalReliability(ReliabilityRestriction):
    """
    A totally fault-free system.

    Neither have any failures occurred nor will they occur.

    *TODO*: Entity failures are not implemented by the simulation.
    """

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        __log_failure_not_implemented__(cls, NotImplementedTypes.entity_failure)
        return (
            network.behavioral_properties.message_loss_indicator is None
        )  # No message loss. Entity failures are not implemented.


class PartialReliability(ReliabilityRestriction):
    """
    No failures will occur.

    Under this restriction, protocols do not need to take failures into account. Note
    that under PartialReliability, failures might have occurred before the execution of a
    computation.

    *TODO*: The simulation does not support partial reliability so this is the same as TotalReliability.
    """

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        __log_failure_not_implemented__(cls, NotImplementedTypes.entity_failure)
        return (
            network.behavioral_properties.message_loss_indicator is None
        )  # The simulation does not support partial reliability (yet)


class GuaranteedDelivery(ReliabilityRestriction):
    """
    Any message that is sent will be received with its content uncorrupted.

    *TODO*: Implement optional message corruption. this only checks for message loss.
    """

    @classmethod
    def check(cls, network: "NetworkType") -> bool:
        __log_failure_not_implemented__(cls, NotImplementedTypes.message_corruption)
        return network.behavioral_properties.message_loss_indicator is None  # Check only no message loss (for now)



from pydistsim.algorithm import NodeAlgorithm, StatusValues
from pydistsim.message import Message
from pydistsim.restrictions.communication import BidirectionalLinks
from pydistsim.restrictions.knowledge import InitialDistinctValues
from pydistsim.restrictions.reliability import TotalReliability
from pydistsim.restrictions.topological import Connectivity


class YoYo(NodeAlgorithm):
    default_params = {
        "neighborsKey": "Neighbors",
        "inNeighborsKey": "InNeighbors",
        "outNeighborsKey": "OutNeighbors",
    }

    class Status(StatusValues):
        INITIATOR = "INITIATOR"
        IDLE = "IDLE"
        SOURCE = "SOURCE"
        INTERMEDIATE = "INTERMEDIATE"
        SINK = "SINK"
        PRUNED = "PRUNED"
        LEADER = "LEADER"

    S_init = (Status.INITIATOR, Status.IDLE)
    S_term = (Status.LEADER, Status.PRUNED)

    algorithm_restrictions = (
        BidirectionalLinks,
        TotalReliability,
        Connectivity,
        InitialDistinctValues,
    )

    # Store assigned id (assigned in SetupYoYo)
    ID_KEY = InitialDistinctValues.KEY

    # Store received ids, I'll use a dict {id_value: [source_nodes]}
    RECEIVED_IDS_KEY = "received_ids"

    # Store received ids that were received while waiting for responses,
    # I'll use a dict {id_value: [source_nodes]}
    RECEIVED_IDS_WHILE_WAITING_RESPONSE_KEY = "received_ids_while_waiting"

    # Store received responses,
    # I'll use a dict {response_value: [source_nodes]}
    RECEIVED_RESPONSES_KEY = "received_responses"

    # Store nodes that requested pruning, list [source_nodes]
    REQUESTED_PRUNING_KEY = "requested_pruning"

    # Store number of sent ids
    SENT_IDS_KEY = "sent_ids"

    # This is sent when a PRUNE is requested
    PRUNE_REQUEST = "prune"

    def initializer(self):
        InitialDistinctValues.apply(self.network)

        for node in self.network.nodes():
            node.memory[self.inNeighborsKey] = []
            node.memory[self.outNeighborsKey] = []

            node.status = self.Status.INITIATOR

            node.memory[self.RECEIVED_IDS_KEY] = {}
            node.memory[self.RECEIVED_RESPONSES_KEY] = {}
            node.memory[self.REQUESTED_PRUNING_KEY] = []
            node.memory[self.SENT_IDS_KEY] = 0
            node.memory[self.RECEIVED_IDS_WHILE_WAITING_RESPONSE_KEY] = {}

            node.push_to_inbox(Message(meta_header=NodeAlgorithm.INI, destination=node))

    def invert_edges(self, node, nodes_to_process, invert_from):
        for node_to_process in nodes_to_process:
            if invert_from == "outNeighbors":
                node.memory[self.outNeighborsKey].remove(node_to_process)
                node.memory[self.inNeighborsKey].append(node_to_process)
            elif invert_from == "inNeighbors":
                node.memory[self.inNeighborsKey].remove(node_to_process)
                node.memory[self.outNeighborsKey].append(node_to_process)

    def prune_nodes(self, node, nodes_to_process, prune_from):
        for node_to_process in nodes_to_process:
            if prune_from == "outNeighbors":
                node.memory[self.outNeighborsKey].remove(node_to_process)
            elif prune_from == "inNeighbors":
                node.memory[self.inNeighborsKey].remove(node_to_process)

    def send_responses(self, node, no_received=False):
        # Find min id
        min_id = min(node.memory[self.RECEIVED_IDS_KEY])
        no_nodes = []
        prune_nodes = []

        received_ids = node.memory[self.RECEIVED_IDS_KEY]
        for received_id in received_ids:
            if no_received:
                # Send NO to all node that sent an id
                self.send(
                    node,
                    destination=received_ids[received_id],
                    header="response",
                    data=(False,),
                )
                no_nodes.extend(received_ids[received_id])

            elif received_id == min_id:
                # Send YES responses to all inNeighbors that send min_id

                if len(received_ids) == 1 and len(node.memory[self.outNeighborsKey]) == 0:
                    # If node received only min_id and has no outNeighbors
                    # remaining after inverting and pruning
                    # it will become a LEAF SINK,
                    # so send a PRUNE request with the YES response as well
                    self.send(
                        node,
                        destination=received_ids[received_id][0],
                        header="response",
                        data=(True, self.PRUNE_REQUEST),
                    )
                    prune_nodes.append(received_ids[received_id][0])
                else:
                    self.send(
                        node,
                        destination=received_ids[received_id][0],
                        header="response",
                        data=(True,),
                    )

                # Send PRUNE request to extra nodes that sent min_id
                # and add them to prune_nodes to be pruned
                self.send(
                    node,
                    destination=received_ids[received_id][1:],
                    header="response",
                    data=(True, self.PRUNE_REQUEST),
                )
                prune_nodes.extend(received_ids[received_id][1:])

            else:
                # Send NO responses to all inNeighbors that didn't
                # send min_id and add them to no_nodes to be inverted
                self.send(
                    node,
                    destination=received_ids[received_id],
                    header="response",
                    data=(False,),
                )
                no_nodes.extend(received_ids[received_id])

        return no_nodes, prune_nodes

    def change_status(self, node):
        if node.status == self.Status.SOURCE:
            if len(node.memory[self.inNeighborsKey]) == 0:
                if len(node.memory[self.outNeighborsKey]) == 0:
                    node.status = self.Status.LEADER
            else:
                if len(node.memory[self.outNeighborsKey]) > 0:
                    node.status = self.Status.INTERMEDIATE
                else:
                    node.status = self.Status.SINK

        elif node.status == self.Status.INTERMEDIATE:
            if len(node.memory[self.outNeighborsKey]) == 0:
                if len(node.memory[self.inNeighborsKey]) == 0:
                    node.status = self.Status.PRUNED

        elif node.status == self.Status.SINK:
            if len(node.memory[self.inNeighborsKey]) == 0:
                node.status = self.Status.PRUNED
            else:
                node.status = self.Status.INTERMEDIATE

        elif node.status == self.Status.IDLE:
            if node.memory[self.inNeighborsKey]:
                node.status = self.Status.SINK
                if node.memory[self.outNeighborsKey]:
                    node.status = self.Status.INTERMEDIATE
            else:
                node.status = self.Status.SOURCE

        self.do(node)

    def receive_id(self, node, message):
        if node.memory[self.SENT_IDS_KEY]:
            ids = node.memory[self.RECEIVED_IDS_WHILE_WAITING_RESPONSE_KEY]

            if message.data in ids:
                ids[message.data].append(message.source)
            else:
                ids[message.data] = [message.source]
        else:
            ids = node.memory[self.RECEIVED_IDS_KEY]

            if message.data in ids:
                ids[message.data].append(message.source)
            else:
                ids[message.data] = [message.source]

    def receive_response(self, node, message):
        responses = node.memory[self.RECEIVED_RESPONSES_KEY]
        response = message.data[0]

        if response in responses:
            responses[response].append(message.source)
        else:
            responses[response] = [message.source]

        if self.PRUNE_REQUEST in message.data:
            node.memory[self.REQUESTED_PRUNING_KEY].append(message.source)

    def do_source(self, node):
        if node.memory[self.SENT_IDS_KEY] == 0:
            node.memory[self.RECEIVED_RESPONSES_KEY] = {}
            node.memory[self.REQUESTED_PRUNING_KEY] = []

            self.send(
                node,
                destination=node.memory[self.outNeighborsKey],
                header="id",
                data=node.memory[self.ID_KEY],
            )

            node.memory[self.SENT_IDS_KEY] = len(node.memory[self.outNeighborsKey])

        else:
            responses = node.memory[self.RECEIVED_RESPONSES_KEY]

            # If responses received for all sent ids handle them
            num_of_responses = sum([len(sources) for sources in list(responses.values())])

            if num_of_responses >= node.memory[self.SENT_IDS_KEY]:
                if False in responses:
                    nodes_to_invert = responses[False]
                else:
                    nodes_to_invert = []

                nodes_to_prune = node.memory[self.REQUESTED_PRUNING_KEY]

                # Invert edges
                self.invert_edges(node, nodes_to_invert, "outNeighbors")

                # Prune nodes
                self.prune_nodes(node, nodes_to_prune, "outNeighbors")

                node.memory[self.RECEIVED_RESPONSES_KEY] = {}
                node.memory[self.REQUESTED_PRUNING_KEY] = []
                node.memory[self.SENT_IDS_KEY] = 0
                node.memory[self.RECEIVED_IDS_KEY] = node.memory[self.RECEIVED_IDS_WHILE_WAITING_RESPONSE_KEY]
                node.memory[self.RECEIVED_IDS_WHILE_WAITING_RESPONSE_KEY] = {}

                # End iteration and change status if needed
                self.change_status(node)

    def do_intermediate(self, node):
        if node.memory[self.SENT_IDS_KEY] == 0:
            ids = node.memory[self.RECEIVED_IDS_KEY]

            # If ids received from all inNeighbors handle them
            num_of_ids = sum([len(sources) for sources in list(ids.values())])

            if num_of_ids >= len(node.memory[self.inNeighborsKey]):
                node.memory[self.RECEIVED_RESPONSES_KEY] = {}
                node.memory[self.REQUESTED_PRUNING_KEY] = []

                # Find min id
                min_id = min(ids)

                # Forward min id to outNeighbors
                self.send(
                    node,
                    destination=node.memory[self.outNeighborsKey],
                    header="id",
                    data=min_id,
                )

                node.memory[self.SENT_IDS_KEY] = len(node.memory[self.outNeighborsKey])

        else:
            responses = node.memory[self.RECEIVED_RESPONSES_KEY]

            # If responses received for all sent ids handle them
            num_of_responses = sum([len(sources) for sources in list(responses.values())])

            if num_of_responses >= node.memory[self.SENT_IDS_KEY]:
                if False in responses:
                    nodes_to_invert = responses[False]
                    no_received = True
                else:
                    nodes_to_invert = []
                    no_received = False

                nodes_to_prune = node.memory[self.REQUESTED_PRUNING_KEY]

                # Invert edges
                self.invert_edges(node, nodes_to_invert, "outNeighbors")

                # Prune nodes
                self.prune_nodes(node, nodes_to_prune, "outNeighbors")

                # Add edges to invert from NO responses sent
                # Add nodes to prune from PRUNE requests sent
                no_response_nodes, prune_nodes = self.send_responses(node, no_received=no_received)

                # Invert edges
                self.invert_edges(node, no_response_nodes, "inNeighbors")

                # Prune nodes
                self.prune_nodes(node, prune_nodes, "inNeighbors")

                node.memory[self.RECEIVED_RESPONSES_KEY] = {}
                node.memory[self.REQUESTED_PRUNING_KEY] = []
                node.memory[self.RECEIVED_IDS_KEY] = node.memory[self.RECEIVED_IDS_WHILE_WAITING_RESPONSE_KEY]
                node.memory[self.RECEIVED_IDS_WHILE_WAITING_RESPONSE_KEY] = {}
                node.memory[self.SENT_IDS_KEY] = 0

                # End iteration and change status if needed
                self.change_status(node)

    def do_sink(self, node):
        ids = node.memory[self.RECEIVED_IDS_KEY]

        # If ids received from all inNeighbors handle them
        num_of_ids = sum([len(sources) for sources in list(ids.values())])

        if num_of_ids >= len(node.memory[self.inNeighborsKey]):
            no_response_nodes, prune_nodes = self.send_responses(node)

            # Invert edges
            self.invert_edges(node, no_response_nodes, "inNeighbors")

            # Prune nodes
            self.prune_nodes(node, prune_nodes, "inNeighbors")

            node.memory[self.RECEIVED_IDS_KEY] = node.memory[self.RECEIVED_IDS_WHILE_WAITING_RESPONSE_KEY]
            node.memory[self.RECEIVED_IDS_WHILE_WAITING_RESPONSE_KEY] = {}
            node.memory[self.SENT_IDS_KEY] = 0

            # End iteration and change status if needed
            self.change_status(node)

    def do(self, node):
        if node.status == self.Status.SOURCE:
            self.do_source(node)
        elif node.status == self.Status.INTERMEDIATE:
            self.do_intermediate(node)
        elif node.status == self.Status.SINK:
            self.do_sink(node)

    @Status.INITIATOR
    def spontaneously(self, node, message):
        # Special case. Only one node in graph
        # If node has no neighbors set its status to LEADER
        if not node.neighbors():
            node.status = self.Status.LEADER
            return

        self.send(
            node,
            header="init_id",
            data=node.memory[self.ID_KEY],
            destination=node.neighbors(),
        )

        node.status = self.Status.IDLE

    @Status.IDLE
    def receiving(self, node, message):
        if message.header == "init_id":
            if message.data < node.memory[self.ID_KEY]:
                node.memory[self.inNeighborsKey].append(message.source)
            else:
                node.memory[self.outNeighborsKey].append(message.source)

            num_of_in_neighbors = len(node.memory[self.inNeighborsKey])
            num_of_out_neighbors = len(node.memory[self.outNeighborsKey])

            if num_of_in_neighbors + num_of_out_neighbors >= len(
                node.neighbors()
            ):  # if all neighbors have sent their ids
                self.change_status(node)

        elif message.header == "id":
            self.receive_id(node, message)

    @Status.SOURCE
    def receiving(self, node, message):
        if message.header == "response":
            self.receive_response(node, message)

        elif message.header == "id":
            self.receive_id(node, message)

        self.do_source(node)

    @Status.INTERMEDIATE
    def receiving(self, node, message):
        if message.header == "id":
            self.receive_id(node, message)

        elif message.header == "response":
            self.receive_response(node, message)

        self.do_intermediate(node)

    @Status.SINK
    def receiving(self, node, message):
        if message.header == "id":
            self.receive_id(node, message)

        elif message.header == "response":
            self.receive_response(node, message)

        self.do_sink(node)

    @Status.PRUNED
    def default(self, node, message):
        pass

    @Status.LEADER
    def default(self, node, message):
        pass



"""
This module contains the restrictions that can be applied or checked in the simulation environment.
"""

# flake8: noqa: F401

from pydistsim.restrictions.base_restriction import (
    ApplicableRestriction,
    CheckableRestriction,
    Restriction,
)



from functools import cached_property
from random import randint
from typing import TYPE_CHECKING, Any, Union

from pydistsim.logging import logger

if TYPE_CHECKING:
    from pydistsim.network import Node
    from pydistsim.network.network import NetworkType


class _NodeWrapper:
    """
    Wrapper class for a node that with controlled access to its attributes.

    Access control is done by defining the attributes that can be accessed in the :attr:`accessible_get` and
    :attr:`accessible_set` class attributes.

    New attributes will be kept only in the wrapper object, while the base node object will remain unchanged except for
    the attributes that are allowed to be changed.
    """

    accessible_get = ()
    "Attributes that can be 'read' from the node base object."

    accessible_set = ()
    "Attributes that can be 'read' or 'written' to the node base object."

    def __init__(self, node: "Node", manager: "WrapperManager", **configs):
        self._node = node
        self._manager = manager
        self._configs = configs

    def __getattr__(self, item):
        if item in self._configs:
            return self._configs[item]
        elif item in self.accessible_get or item in self.accessible_set:
            return getattr(self._node, item)
        raise AttributeError(f"{self.__class__.__name__} object has no attribute {item}")

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.accessible_set:
            setattr(self._node, name, value)
        else:
            super().__setattr__(name, value)

    def __repr__(self):
        return self._node.__repr_str__(self.id)

    def __deepcopy__(self, memo):
        # Do not copy the object, just return the same object
        memo[id(self)] = self
        return self

    def __copy__(self):
        # Do not copy the object, just return the same object
        return self

    def unbox(self) -> "Node":
        return self._node

    @property
    def id(self):
        raise AttributeError("Stub method. Raise AttributeError to trigger the __getattr__ method.")


class NeighborLabel(_NodeWrapper):
    """
    Class that represents a neighbor of a node. It is used to represent the knowledge that a node has about its
    neighbors.
    """

    def __repr__(self):
        return f"Neighbor(label={self.id})"

    @property
    def id(self):
        logger.warning(
            "Neighbor's id do not correspond to the real id of the node. It can be used to distinguish "
            "neighbors from each other."
        )
        return super().id


class NodeAccess(_NodeWrapper):
    """
    Class used to control the access to a node's attributes.

    For full node access, use the :meth:`unbox` method. Be aware that such access may break the knowledge restrictions
    of the algorithm.
    """

    accessible_get = (
        "status",
        "memory",
        "clock",
    )

    accessible_set = (
        "status",
        "memory",
    )

    def neighbors(self) -> set["NeighborLabel"]:
        """
        Get the out-neighbors of the node.

        Keep in mind that `neighbor.id` is only a label for the neighbor, it's local to the node and it's only
        guaranteed to be unique among the neighbors of the node.

        :return: The out-neighbors of the node.
        :rtype: set[NeighborAccess]
        """

        return set(self.__out_neighbors_dict.values())

    def in_neighbors(self) -> set["NeighborLabel"]:
        """
        Get the in-neighbors of the node. If the network is not directed, the in-neighbors are the same as the
        out-neighbors.

        Keep in mind that `neighbor.id` is only a label for the neighbor, it's local to the node and it's only
        guaranteed to be unique among the neighbors of the node.

        :return: The in-neighbors of the node.
        :rtype: set[NeighborAccess]
        """

        return set(self.__in_neighbors_dict.values())

    out_neighbors = neighbors
    "Alias for out_neighbors."

    @property
    def id(self):
        """
        Get the id of the node. If the node does not have an id in memory, a random id will be generated.
        It is not guaranteed that the id will be unique among all nodes in the network.

        Since the id is a read-only attribute, it is cached to avoid generating a new id every time it is accessed.
        """
        if "id" not in self._node.memory:
            logger.warning(
                "Node's id do not correspond to the real id of the node. It can't be used to distinguish nodes from "
                "each other as it is not unique unless restriction InitialDistinctValues is applied (or each node has a "
                "unique id set in memory)."
            )
            return self._rand_id

        return self._node.memory["id"]

    ###### Private methods ######

    @cached_property
    def __out_neighbors_dict(self) -> dict["Node", "NeighborLabel"]:
        return self._manager.node_out_neighbor_labels[self._node]

    @cached_property
    def __in_neighbors_dict(self) -> dict["Node", "NeighborLabel"]:
        if not self._node.network.is_directed():
            return self.__out_neighbors_dict  # Bidirectional links, same neighbors and same labels
        else:
            return self._manager.node_in_neighbor_labels[self._node]

    def _get_out_neighbor_proxy(self, node: "Node") -> "NeighborLabel":
        return self.__out_neighbors_dict[node]

    def _get_in_neighbor_proxy(self, node: "Node") -> "NeighborLabel":
        return self.__in_neighbors_dict[node]

    @cached_property
    def _rand_id(self):
        return randint(0, len(self._node.network))


class SensorNodeAccess(NodeAccess):
    accessible_get = (
        *NodeAccess.accessible_get,  # All attributes from NodeAccess
        "sensors",
        "compositeSensor",
    )


class WrapperManager:
    """
    Node Access class
    -----------------

    A node's own view is a proxy for the node that represents the knowledge of the node.

    `u_own_view` and `v_own_view` are the same object passed as parameters at each algorithm's action:

    .. code-block:: python

        @Status.IDLE
        def receiving(self, node: NodeAccess, message: Message):
                                ↑


    Neighbor Label class
    --------------------

    A label for a neighbor is a proxy for the neighbor that represents the knowledge of the neighbor respect to the
    node.

    Alternatively, `u_label_for_v` and `v_label_for_u` are the objects that are returned when a node asks for its
    neighbors in some way:

    .. code-block:: python

        @Status.IDLE
        def receiving(self, node: NodeAccess, message: Message):
            source = message.source  # The node that sent the message
                ↑

            for neighbor in node.neighbors():
                    ↑
    """

    NODE_ACCESS_TYPE = NodeAccess
    "The class that will be used to wrap the nodes' own view."

    NEIGHBOR_LABEL_TYPE = NeighborLabel
    "The class that will be used to wrap the neighbors' labels."

    def __init__(self, network: "NetworkType"):
        self.network = network
        self.node_access_instances = {}
        self.node_out_neighbor_labels = {}
        self.node_in_neighbor_labels = {}

    def get_node_access(self, node: Union["Node", "NodeAccess"]) -> NodeAccess:
        if isinstance(node, NodeAccess):
            return node

        if node not in self.node_access_instances:
            node_access = self.NODE_ACCESS_TYPE(node, self)
            self.node_access_instances[node] = node_access
            self.node_out_neighbor_labels[node] = self.calculate_out_neighbors_dict(node_access)
            if self.network.is_directed():
                self.node_in_neighbor_labels[node] = self.calculate_in_neighbors_dict(node_access)

        return self.node_access_instances[node]

    def calculate_out_neighbors_dict(self, node_access: NodeAccess) -> dict["Node", "NeighborLabel"]:
        return {
            node: self.NEIGHBOR_LABEL_TYPE(node, self, id=i)
            for i, node in enumerate(self.network.out_neighbors(node_access._node))
        }

    def calculate_in_neighbors_dict(self, node_access: NodeAccess) -> dict["Node", "NeighborLabel"]:
        assert self.network.is_directed(), "The network must be directed to calculate in-neighbors."

        return {
            node: self.NEIGHBOR_LABEL_TYPE(node, id=i)
            for i, node in enumerate(self.network.in_neighbors(node_access._node))
        }

    def edges(self):
        """
        Returns the edges of the network as a tuple of proxies for the nodes and the labels of the neighbors.
        For each edge `(u, v)`, it returns two tuples:
        - `(u_own_view, v_label_for_u)`
        - `(v_own_view, u_label_for_v)`
        """

        for u, v in self.network.edges():
            u_own_view = self.get_node_access(u)
            v_label_for_u = u_own_view._get_out_neighbor_proxy(v)

            v_own_view = self.get_node_access(v)
            u_label_for_v = v_own_view._get_in_neighbor_proxy(u)

            yield (u_own_view, v_label_for_u), (v_own_view, u_label_for_v)

    def nodes(self):
        """
        Returns the proxies for the nodes in the network.
        """

        for node in self.network.nodes():
            yield self.get_node_access(node)

    def nodes_neighbors(self):
        """
        Returns the proxies for the nodes and the labels of it's neighbors.
        """

        for u in self.network.nodes():
            u_node_access = self.get_node_access(u)
            neighbor_labels = u_node_access.out_neighbors()

            yield u_node_access, neighbor_labels



import unittest

from pydistsim import Node
from pydistsim._exceptions import SimulationException
from pydistsim.algorithm import (
    Actions,
    AlgorithmException,
    NetworkAlgorithm,
    NodeAlgorithm,
    StatusValues,
)
from pydistsim.message import Message
from pydistsim.network import NetworkException, NetworkGenerator
from pydistsim.simulation import Simulation
from pydistsim.utils.helpers import first
from pydistsim.utils.testing import PyDistSimTestCase


def set_algorithms(sim, algorithms):
    sim.algorithms = algorithms


class SomeNodeAlgorithm(NodeAlgorithm):
    required_params = ("rp1", "rp2", "rp3")
    default_params = {"dp1": "dp1_value", "dp2": "dp2_value", "dp3": "dp3_value"}


class SomeNetworkAlgorithm(NetworkAlgorithm):
    default_params = {"dp1": "dp1_value", "dp2": "dp2_value", "dp3": "dp3_value"}


class SomeAlgorithmWithInheritance(SomeNodeAlgorithm):
    required_params = ("rp4",)
    default_params = {
        "dp4": "dp4_value",
    }


class SomeAlgorithmWithInheritanceChild(SomeAlgorithmWithInheritance):
    required_params = ("rp5", "rp6")
    default_params = {
        "dp2": "overriden_dp2_value",
    }


def rp_multiple():
    class SomeAlgorithmWhereRpIsRedefined(SomeAlgorithmWithInheritance):
        required_params = ("rp2",)


def dp_is_rp():
    class SomeAlgorithmWhereDpIsInheritedRp(SomeAlgorithmWithInheritance):
        default_params = {
            "rp2": "dp2_value",
        }


def rp_is_dp():
    class SomeAlgorithmWhereRpIsInheritedDp(SomeAlgorithmWithInheritance):
        required_params = ("dp2",)


class TestAlgorithmsSetter(unittest.TestCase):

    def setUp(self):
        net_gen = NetworkGenerator(100)
        self.net = net_gen.generate_random_network()
        self.algorithms_ok = (
            (
                SomeNodeAlgorithm,
                {
                    "rp1": 1,
                    "rp2": 2,
                    "rp3": 3,
                },
            ),
            (SomeNetworkAlgorithm, {}),
            SomeNetworkAlgorithm,
            (SomeAlgorithmWithInheritance, {"rp1": 1, "rp2": 2, "rp3": 3, "rp4": 4}),
            (
                SomeAlgorithmWithInheritanceChild,
                {"rp1": 1, "rp2": 2, "rp3": 3, "rp4": 4, "rp5": 5, "rp6": 6},
            ),
        )
        self.check = [
            # wrong_format
            (
                SimulationException,
                [
                    (SomeNodeAlgorithm, {"rp1": 1, "rp2": 2, "rp3": 3}),
                ],
            ),
            # wrong_base_class
            (SimulationException, ((Node, {}),)),
            # missing_req_params
            (
                AlgorithmException,
                (
                    (
                        SomeNodeAlgorithm,
                        {
                            "rp1": 1,
                        },
                    ),
                ),
            ),
            (
                AlgorithmException,
                (
                    (
                        SomeAlgorithmWithInheritance,
                        {
                            "rp1": 1,
                        },
                    ),
                ),
            ),
        ]

    def test_setter(self):
        """Test different algorithm initialization formats and params."""
        sim = Simulation(self.net)

        set_algorithms(sim, self.algorithms_ok)
        for exc, alg in self.check:
            self.assertRaises(exc, set_algorithms, sim, alg)

    def test_params_inheritance(self):
        """Test default params inheritance algorithm classes."""
        sim = Simulation(self.net)
        sim.algorithms = (
            (
                SomeAlgorithmWithInheritanceChild,
                {"rp1": 1, "rp2": 2, "rp3": 3, "rp4": 4, "rp5": 5, "rp6": 6},
            ),
        )
        self.assertTrue(sim.algorithms[0].dp1 == "dp1_value")
        self.assertTrue(sim.algorithms[0].dp2 == "overriden_dp2_value")
        self.assertTrue(sim.algorithms[0].dp3 == "dp3_value")
        self.assertRaises(AssertionError, rp_multiple)
        self.assertRaises(AssertionError, dp_is_rp)
        self.assertRaises(AssertionError, rp_is_dp)

    def test_default_params(self):
        """Test default params."""
        sim = Simulation(self.net)
        sim.algorithms = (
            (
                SomeNetworkAlgorithm,
                {
                    "dp2": "overriden_dp2_value",
                },
            ),
        )
        self.assertTrue(sim.algorithms[0].dp1 == "dp1_value")
        self.assertTrue(sim.algorithms[0].dp2 == "overriden_dp2_value")
        self.assertTrue(sim.algorithms[0].dp3 == "dp3_value")


class TestStatusValues(unittest.TestCase):
    def setUp(self):
        class Status(StatusValues):
            IDLE = "IDLE"
            DONE = "DONE"

        self.Status = Status

    def test_state_action_names(self):
        """Test if state action names raise exception and correct ones are allowed."""

        @self.Status.IDLE
        def spontaneously(): ...

        @self.Status.IDLE
        def receiving(): ...

        @self.Status.IDLE
        def alarm(): ...

        # Test if wrong name raises exception
        def wrong_name(): ...

        self.assertRaises(AssertionError, self.Status.IDLE, wrong_name)

        # Test if good name with capital letters raises exception
        def ALARM(): ...

        self.assertRaises(AssertionError, self.Status.IDLE, ALARM)

    def test_implements_method(self):
        """Test Status.implements method."""

        @self.Status.DONE
        def alarm(): ...

        # Using correct ActionEnum
        assert self.Status.DONE.implements(Actions.alarm)

        # Using wrong ActionEnum
        assert not self.Status.DONE.implements(Actions.spontaneously)

        # Using string
        assert self.Status.DONE.implements("alarm")

        # Using wrong string
        assert not self.Status.DONE.implements("spontaneously")

        # Using string with capital letters
        assert not self.Status.DONE.implements("ALARM")


class TimerAlgorithm(NodeAlgorithm):
    class Status(StatusValues):
        IDLE = "IDLE"
        DONE = "DONE"

    def initializer(self):
        for node in self.network.nodes():
            node.status = self.Status.IDLE
            self.set_alarm(node, 3, Message(data=f"ALARM{node._internal_id}"))
            self.set_alarm(node, 6, Message(data=f"ALARM{node._internal_id}"))
            node.memory["LAST_ALARM"] = self.set_alarm(node, 9, Message(data=f"ALARM{node._internal_id}"))

    @Status.IDLE
    def alarm(self, node, message):
        node.memory["alarm"] = message.data

        self.disable_alarm(node.memory["LAST_ALARM"])
        self.disable_all_node_alarms(node)
        node.status = self.Status.DONE


class TimerDefaultMessage(TimerAlgorithm):
    def initializer(self):
        for node in self.network.nodes():
            node.status = self.Status.IDLE
            self.set_alarm(node, 3)
            self.set_alarm(node, 6)
            node.memory["LAST_ALARM"] = self.set_alarm(node, 9)


class TimerDefaultMessageDecreasedTime(TimerAlgorithm):

    @TimerAlgorithm.Status.IDLE
    def alarm(self, node, message):
        node.memory["alarm"] = message.data

        self.disable_alarm(node.memory["LAST_ALARM"])
        self.disable_all_node_alarms(node)
        node.status = self.Status.DONE

    @TimerAlgorithm.Status.IDLE
    def spontaneously(self, node, message):
        self.update_alarm_time(node.memory["FIRST_ALARM"], -100)
        self.disable_alarm(node.memory["LAST_ALARM"])

    def initializer(self):
        for node in self.network.nodes():
            node.status = self.Status.IDLE
            node.memory["FIRST_ALARM"] = self.set_alarm(node, 4)
            node.memory["SECOND_ALARM"] = self.set_alarm(node, 6)
            node.memory["LAST_ALARM"] = self.set_alarm(node, 9)
            node.push_to_inbox(Message(meta_header=NodeAlgorithm.INI))


class TestAlarms(unittest.TestCase):

    def test_run_base_algorithm(self):
        def node_data_test(node: "Node"):
            return (node.status == TimerAlgorithm.Status.DONE and node.memory["alarm"] == f"ALARM{node._internal_id}",)

        self.aux_test(TimerAlgorithm, node_data_test)

    def test_run_base_algorithm_default_message(self):
        self.aux_test(
            TimerDefaultMessage,
            lambda node: node.status == TimerAlgorithm.Status.DONE and node.memory["alarm"] is None,
        )

    def aux_test(self, algo_class, data_test):
        self.net = NetworkGenerator(10).generate_random_network()
        sim = Simulation(self.net)
        sim.algorithms = (algo_class,)

        algorithm: NodeAlgorithm = sim.get_current_algorithm()
        some_n = first(self.net.nodes())

        assert sim.is_halted()

        # 1 step for initialization, 3 steps for alarms, 1 step for processing alarm messages
        sim.run(1)

        assert len(algorithm._alarms) == 3 * len(self.net.nodes())
        assert len(some_n.inbox) == 0

        sim.run(1)  # alarm tic

        assert len(algorithm._alarms) == 3 * len(self.net.nodes())
        assert len(some_n.inbox) == 0

        sim.run(1)  # alarm tic

        assert len(algorithm._alarms) == 3 * len(self.net.nodes())
        assert len(some_n.inbox) == 0

        sim.run(1)  # alarm tic

        assert len(algorithm._alarms) == 2 * len(self.net.nodes())
        assert all(len(node.inbox) == 1 for node in self.net.nodes())

        sim.run(1)  # 1 step for processing alarm messages

        assert all([data_test(node) for node in self.net.nodes()])
        assert len(algorithm._alarms) == 0
        assert sim.is_halted()

    def test_update_alarm(self):
        self.net = NetworkGenerator(10).generate_random_network()
        sim = Simulation(self.net)
        sim.algorithms = (TimerDefaultMessageDecreasedTime,)

        algorithm: NodeAlgorithm = sim.get_current_algorithm()
        some_n = first(self.net.nodes())

        assert sim.is_halted()

        # 1 step for initialization, 3 steps for alarms, 1 step for processing alarm messages
        sim.run(1)

        assert len(algorithm._alarms) == 3 * len(self.net.nodes())
        assert len(some_n.inbox) == 1  # INI message
        assert some_n.memory["FIRST_ALARM"].time_left == 4

        sim.run(1)  # alarm tic

        assert len(algorithm._alarms) == 3 * len(self.net.nodes())
        assert len(some_n.inbox) == 1  # INI message
        assert some_n.memory["FIRST_ALARM"].time_left == 3

        sim.run(1)  # alarm tic -> processes INI and disables LAST_ALARM

        assert len(algorithm._alarms) == 2 * len(self.net.nodes())
        assert len(some_n.inbox) == 0
        assert some_n.memory["FIRST_ALARM"].time_left == -98  # -100 + 2

        sim.run(1)  # alarm tic -> FIRST_ALARM triggered

        assert len(algorithm._alarms) == len(self.net.nodes())  # Only SECOND_ALARM left
        assert all(len(node.inbox) == 1 for node in self.net.nodes())  # 1 alarm message (generated by FIRST_ALARM)

        sim.run(1)  # 1 step for processing alarm messages, disables SECOND_ALARM

        assert len(algorithm._alarms) == 0
        assert sim.is_halted()



import unittest
from inspect import isclass
from math import inf as Inf

from pydistsim.demo_algorithms.readsensors import ReadSensors
from pydistsim.network import NetworkGenerator, NetworkGeneratorException, UdgRangeType
from pydistsim.network.environment import Environment2D
from pydistsim.network.network import NetworkType
from pydistsim.utils.testing import PyDistSimTestCase


class TestNetworkGeneration(PyDistSimTestCase):

    def setUp(self):
        # Raises NetworkGeneratorException
        # returns None
        # else expected network/node properties dictionary
        env = Environment2D(shape=(600, 600))
        rangeType = UdgRangeType(env)
        algorithms = (ReadSensors,)
        self.in_out = [
            (
                "default N_COUNT and COMM_RANGE and ENVIRONMENT should be compatible",
                {
                    "n_count": None,
                    "n_min": 0,
                    "n_max": Inf,
                    "enforce_connected": True,
                    "environment": None,
                    "degree": None,
                    "comm_range": None,
                },
                {"count": list(range(100, 1001))},
            ),
            (
                "regular default params",
                {
                    "n_count": 100,
                    "n_min": 0,
                    "n_max": Inf,
                    "enforce_connected": True,
                    "environment": env,
                    "degree": None,
                    "comm_range": 100,
                },
                {"count": list(range(100, 1001))},
            ),
            ############## connected True degree False
            (
                "increase node number",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": Inf,
                    "enforce_connected": True,
                    "environment": env,
                    "degree": None,
                    "comm_range": 100,
                },
                {"count": list(range(11, 301))},
            ),
            (
                "increase commRange",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": 10,
                    "enforce_connected": True,
                    "environment": env,
                    "degree": None,
                    "comm_range": None,
                },
                {"count": 10},
            ),
            (
                "decrease commRange",
                {
                    "n_count": 10,
                    "n_min": 10,
                    "n_max": 10,
                    "enforce_connected": True,
                    "environment": env,
                    "degree": 0,
                    "comm_range": None,
                },
                {"count": 10},
            ),
            ############## connected True degree True
            (
                "increase node number",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": 200,
                    "enforce_connected": True,
                    "environment": env,
                    "degree": 11,
                    "comm_range": 100,
                },
                {"count": list(range(10, 201))},
            ),
            (
                "increase commRange",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": 10,
                    "enforce_connected": True,
                    "environment": env,
                    "degree": 9,
                    "degree_tolerance": 0.01,
                    "comm_range": None,
                },
                {"count": 10},
            ),
            (
                "low degree with connected, alternating directions problem",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": 10,
                    "enforce_connected": True,
                    "environment": env,
                    "degree": 3,
                    "comm_range": 30,
                },
                None,
            ),
            ############## connected False degree True
            (
                "increase node number",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": 200,
                    "enforce_connected": False,
                    "environment": env,
                    "degree": 8,
                    "comm_range": 100,
                },
                {"count": list(range(10, 201))},
            ),
            (
                "increase commRange",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": 200,
                    "enforce_connected": False,
                    "environment": env,
                    "degree": 11,
                    "comm_range": None,
                },
                {"count": list(range(10, 201))},
            ),
            (
                "low degree",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": 100,
                    "enforce_connected": False,
                    "environment": env,
                    "degree": 3,
                    "comm_range": 100,
                },
                {"count": list(range(10, 101))},
            ),
            (
                "degree too high for node number",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": 10,
                    "enforce_connected": False,
                    "environment": env,
                    "degree": 10,
                    "comm_range": None,
                },
                NetworkGeneratorException,
            ),
            (
                "",
                {
                    "n_count": 11,
                    "n_min": 0,
                    "n_max": 10,
                    "enforce_connected": False,
                    "environment": env,
                    "degree": None,
                    "comm_range": None,
                },
                NetworkGeneratorException,
            ),
            (
                "",
                {
                    "n_count": 9,
                    "n_min": 10,
                    "n_max": 10,
                    "enforce_connected": False,
                    "environment": env,
                    "degree": None,
                    "comm_range": None,
                },
                NetworkGeneratorException,
            ),
            ############## connected False degree False - no need for modifying initially created network
            # also remove environment from kwargs to test default and change comm_range to commRange
            (
                "",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": 100,
                    "enforce_connected": False,
                    "degree": None,
                    "commRange": 100,
                },
                {"count": 10},
            ),
            (
                "",
                {
                    "n_count": 20,
                    "n_min": 0,
                    "n_max": 100,
                    "enforce_connected": False,
                    "degree": None,
                    "commRange": None,
                },
                {"count": 20},
            ),
            (
                "",
                {
                    "n_count": 30,
                    "n_min": 0,
                    "n_max": 100,
                    "enforce_connected": False,
                    "degree": None,
                    "commRange": 30,
                },
                {"count": 30},
            ),
            (
                "Check sensors and algorithms",
                {
                    "n_count": 10,
                    "n_min": 0,
                    "n_max": 100,
                    "enforce_connected": False,
                    "rangeType": rangeType,
                    "algorithms": algorithms,
                    "commRange": 100,
                },
                {
                    "count": 10,
                    "rangeType": rangeType,
                    "algorithms": algorithms,
                    "commRange": 100,
                },
            ),
        ]

    def test_random_generation(self):
        """Test different random generation parameters"""
        for comment, input, output in self.in_out:
            for directed in (True, False):
                with self.subTest(input=input, output=output, directed=directed):
                    print(f"{comment=}")
                    print(f"{input=}\n{output=}")
                    if isclass(output) and issubclass(output, Exception):
                        self.assertRaises(output, NetworkGenerator, **input)
                        continue
                    net_gen = NetworkGenerator(directed=directed, **input)
                    if output is None:
                        self.assertEqual(None, net_gen.generate_random_network(max_steps=500))
                    elif isinstance(output, dict):
                        net = net_gen.generate_random_network(max_steps=2000)
                        assert isinstance(net, NetworkType)
                        assert directed == net.is_directed()
                        net.validate_params(output)


class TestNotConnectedNetworkGeneration(unittest.TestCase):
    def setUp(self) -> None:
        self.gen = NetworkGenerator(
            n_count=10,
            n_min=10,
            n_max=10,
            degree=0.01,
            degree_tolerance=0.2,
            enforce_connected=False,
        )

    def test_generate_not_connected_net(self):
        for directed in (True, False):
            with self.subTest(directed=directed):
                self.gen.directed = directed
                net = self.gen.generate_random_network()
                assert not net.is_connected()
                assert len(net) == 10
                assert abs(net.avg_degree()) <= 0.2
                assert net.is_directed() == directed


class TestOtherGenerators(unittest.TestCase):
    def setUp(self) -> None:
        self.gen1 = NetworkGenerator(n_count=10, n_min=0, n_max=100, enforce_connected=True)
        self.gen2 = NetworkGenerator(n_count=10, n_min=10, n_max=10, enforce_connected=True)
        self.gen3 = NetworkGenerator(n_count=99, n_min=99, n_max=99, enforce_connected=True)
        self.gen4 = NetworkGenerator(
            n_count=10,
            n_min=10,
            n_max=10,
            degree=0,
            degree_tolerance=0.2,
            enforce_connected=False,
        )

    def test_generate_neighborhood(self):
        net1 = self.gen1.generate_neigborhood_network()

        assert len(net1) >= 10 and len(net1) <= 100
        assert net1.is_connected()

        net2 = self.gen2.generate_neigborhood_network()

        assert len(net2) == 10
        assert net2.is_connected()

        net3 = self.gen3.generate_neigborhood_network()

        assert len(net3) == 99
        assert net3.is_connected()

        # Dont test for the connectedness and degree for this network
        # as it is not enforced

    def test_homogeneous_network(self):
        net1 = self.gen1.generate_homogeneous_network()

        assert len(net1) >= 10 and len(net1) <= 100
        assert net1.is_connected()

        net2 = self.gen2.generate_homogeneous_network()

        assert len(net2) == 10
        assert net2.is_connected()

        net3 = self.gen3.generate_homogeneous_network()

        assert len(net3) == 99
        assert net3.is_connected()

        net4 = self.gen4.generate_homogeneous_network()

        assert len(net4) == 10
        assert not net4.is_connected()
        assert abs(net4.avg_degree()) <= 0.2



import unittest

import pytest
from networkx import is_directed_acyclic_graph, is_weakly_connected

from pydistsim.algorithm import NodeAlgorithm
from pydistsim.network import (
    BidirectionalNetwork,
    BidirectionalRangeNetwork,
    CompleteRangeType,
    Network,
    NetworkException,
    NetworkType,
    Node,
    RangeNetwork,
    RangeType,
)
from pydistsim.network.environment import Environment2D
from pydistsim.utils.testing import PyDistSimTestCase
from pydistsim.utils.tree import check_tree_key


class TestDirectedNetwork(PyDistSimTestCase):
    nw_class = RangeNetwork

    def setUp(self):
        env = Environment2D()
        self.net = self.nw_class(rangeType=CompleteRangeType(env))
        self.net.environment.image[22, 22] = 0
        self.node1 = self.net.add_node(pos=[22.8, 21.8])
        self.node1.memory["test"] = 1
        self.node2 = self.net.add_node(pos=[21.9, 22.9])
        self.node2.memory["test"] = 2
        self.node3 = self.net.add_node(pos=[21.7, 21.7])
        self.node3.memory["test"] = 3
        self.node4 = self.net.add_node(pos=[23.7, 23.7])
        self.node4.memory["test"] = 4

        self.other_net = self.nw_class(rangeType=CompleteRangeType(env))
        self.node_in_other_net = self.other_net.add_node()

    def test_nodes(self):
        """Make sure the nodes are added."""
        assert isinstance(self.node1, Node)
        assert len(self.net.nodes()) == 4
        if isinstance(self.net.environment, Environment2D):
            assert self.net.environment._image.shape == (600, 600), "incorrect default size"

        assert isinstance(self.net.rangeType, RangeType)

        assert all(node.network == self.net for node in self.net.nodes())

    def test_visibility(self):
        """
        Pixel 22,22 is not space so node1 and node2 should not be visible
        but node3 is visible.
        """
        assert not self.net.environment.are_visible(self.net.pos[self.node1], self.net.pos[self.node2])

        assert self.net.environment.are_visible(self.net.pos[self.node2], self.net.pos[self.node3])

    def test_subnetwork(self):
        """Test subnetwork creation."""
        subnetwork: "NetworkType" = self.net.subnetwork([self.node1, self.node2])

        assert len(subnetwork.nodes()) == 2
        assert isinstance(subnetwork, type(self.net))

        for og_node in [self.node1, self.node2]:
            sn_node = [n for n in subnetwork if n.memory["test"] == og_node.memory["test"]][0]
            assert sn_node.network == subnetwork
            assert isinstance(sn_node, type(og_node))
            assert subnetwork.labels[sn_node] == str(sn_node._internal_id)
            assert subnetwork.ori[sn_node] == self.net.ori[og_node]
            assert all(subnetwork.pos[sn_node] == self.net.pos[og_node])

    def test_nodes_sorted(self):
        """Test sorting of nodes."""
        assert self.net.nodes_sorted() == [
            self.node1,
            self.node2,
            self.node3,
            self.node4,
        ]

    def test_remove_node(self):
        """Test node removal."""
        pos = self.net.pos[self.node1]
        original_len = len(self.net.nodes())

        print(f"Nodes ids pre-remove: {[node._internal_id for node in self.net.nodes()]}")
        self.net.remove_node(self.node1)
        print(f"Nodes ids post-remove: {[node._internal_id for node in self.net.nodes()]}")
        assert len(self.net.nodes()) == original_len - 1
        assert self.node1 not in self.net.nodes()

        with self.assertRaises(NetworkException):
            self.net.remove_node(self.node1)

        self.net.add_node(self.node1, pos=pos)

    def test_add_node_error_in_other_net(self):
        """Test adding node to network."""
        with self.assertRaises(NetworkException):
            self.net.add_node(self.node_in_other_net)

    def test_get_dic(self):
        """Test getting dictionary representation of the network."""
        dic = self.net.get_dic()
        assert "nodes" in dic
        assert "edges" in dic

    def test_get_tree_net(self):
        """Test getting tree representation of the network."""
        tree_key = "T_KEY"
        keepMemKey = "TEST_KEEP_MEM"

        self.node1.memory[tree_key] = {
            "parent": None,
            "children": [self.node2, self.node3],
        }
        self.node2.memory[tree_key] = {"parent": self.node1, "children": []}
        self.node2.memory[keepMemKey] = {"test": 1}
        self.node3.memory[tree_key] = {"parent": self.node1, "children": []}
        self.node4.memory[tree_key] = {"parent": None, "children": []}

        tree = self.net.get_tree_net(tree_key)

        assert isinstance(tree, self.nw_class)
        assert len(tree.nodes()) == 3
        assert len(self.net.nodes()) == 4
        assert tree.is_connected()
        assert (node.network == tree for node in tree.nodes())

        assert keepMemKey in self.node2.memory, "Memory should be kept"
        check_tree_key(self.net, tree_key)

    @pytest.mark.filterwarnings("ignore:No data for colormapping.*")
    def test_get_fig_runs(self):
        """Test getting figure of the network."""
        assert self.net.get_fig() is not None

        assert self.other_net.get_fig() is not None


class TestUnDirectedNetwork(TestDirectedNetwork):
    nw_class = BidirectionalRangeNetwork
